import autogalaxy as ag
from autoarray.operators.inversion import inversions

import numpy as np
import pytest

from test_autogalaxy.mock.mock_profiles import MockLightProfile


class TestFitImaging:
    class TestLikelihood:
        def test__1x2_image__no_psf_blurring__plane_fits_data_with_chi_sq_5(self):
            # The image plane image generated by the galaxy is [1.0, 1.0]

            # Thus the chi squared is 4.0**2.0 + 3.0**2.0 = 25.0

            psf = ag.Kernel.manual_2d(
                array=(np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]])),
                pixel_scales=1.0,
            )

            imaging = ag.Imaging(
                image=5.0 * ag.Array.ones(shape_2d=(3, 4)),
                psf=psf,
                noise_map=ag.Array.ones(shape_2d=(3, 4)),
            )
            imaging.image[6] = 4.0

            mask = ag.Mask.manual(
                mask_2d=np.array(
                    [
                        [True, True, True, True],
                        [True, False, False, True],
                        [True, True, True, True],
                    ]
                ),
                pixel_scales=1.0,
                sub_size=1,
            )

            masked_imaging_7x7 = ag.MaskedImaging(imaging=imaging, mask=mask)

            # Setup as a ray trace instance, using a light profile for the lens

            g0 = ag.Galaxy(
                redshift=0.5, light_profile=MockLightProfile(value=1.0, size=2)
            )
            plane = ag.Plane(galaxies=[g0])

            fit = ag.FitImaging(masked_imaging=masked_imaging_7x7, plane=plane)

            assert (
                fit.mask
                == np.array(
                    [
                        [True, True, True, True],
                        [True, False, False, True],
                        [True, True, True, True],
                    ]
                )
            ).all()

            assert (
                fit.image.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 5.0, 4.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.noise_map.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 1.0, 1.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.model_image.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 1.0, 1.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.residual_map.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 4.0, 3.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.normalized_residual_map.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 4.0, 3.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.chi_squared_map.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 16.0, 9.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert fit.chi_squared == 25.0
            assert fit.reduced_chi_squared == 25.0 / 2.0
            assert fit.noise_normalization == (2.0 * np.log(2 * np.pi * 1.0 ** 2.0))
            assert fit.log_likelihood == -0.5 * (
                25.0 + 2.0 * np.log(2 * np.pi * 1.0 ** 2.0)
            )

        def test__1x2_image__include_psf_blurring__plane_fits_data_with_chi_sq_4(self):
            # This PSF changes the blurred image plane image from [1.0, 1.0] to [1.0, 5.0]

            # Thus, the chi squared is 4.0**2.0 + 0.0**2.0 = 16.0

            psf = ag.Kernel.manual_2d(
                array=(np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 3.0], [0.0, 0.0, 0.0]])),
                pixel_scales=1.0,
                renormalize=False,
            )

            imaging = ag.Imaging(
                image=5.0 * ag.Array.ones(shape_2d=(3, 4)),
                psf=psf,
                noise_map=ag.Array.ones(shape_2d=(3, 4)),
            )
            imaging.image[6] = 4.0

            mask = ag.Mask.manual(
                mask_2d=np.array(
                    [
                        [True, True, True, True],
                        [True, False, False, True],
                        [True, True, True, True],
                    ]
                ),
                pixel_scales=1.0,
                sub_size=1,
            )

            masked_imaging_7x7 = ag.MaskedImaging(
                imaging=imaging, mask=mask, renormalize_psf=False
            )

            # Setup as a ray trace instance, using a light profile for the lens

            g0 = ag.Galaxy(
                redshift=0.5, light_profile=MockLightProfile(value=1.0, size=2)
            )
            plane = ag.Plane(galaxies=[g0])

            fit = ag.FitImaging(masked_imaging=masked_imaging_7x7, plane=plane)

            assert (
                fit.mask
                == np.array(
                    [
                        [True, True, True, True],
                        [True, False, False, True],
                        [True, True, True, True],
                    ]
                )
            ).all()

            assert (
                fit.image.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 5.0, 4.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.noise_map.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 1.0, 1.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.model_image.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 1.0, 4.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.residual_map.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 4.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.normalized_residual_map.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 4.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.chi_squared_map.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 16.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert fit.chi_squared == 16.0
            assert fit.reduced_chi_squared == 16.0 / 2.0
            assert fit.noise_normalization == (2.0 * np.log(2 * np.pi * 1.0 ** 2.0))
            assert fit.log_likelihood == -0.5 * (
                16.0 + 2.0 * np.log(2 * np.pi * 1.0 ** 2.0)
            )

        def test_hyper_galaxy_changes_noise_above_from_1_to_2__reflected_in_likelihood(
            self
        ):
            # This PSF changes the blurred image plane image from [1.0, 1.0] to [1.0, 5.0]

            # Thus, the chi squared is 4.0**2.0 + 0.0**2.0 = 16.0

            # The hyper_galaxies galaxy increases the noise in both pixels by 1.0, to 2.0.

            # This reduces the chi squared to 2.0 instead of 4.0

            psf = ag.Kernel.manual_2d(
                array=(np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 3.0], [0.0, 0.0, 0.0]])),
                pixel_scales=1.0,
            )

            imaging = ag.Imaging(
                image=5.0 * ag.Array.ones(shape_2d=(3, 4)),
                psf=psf,
                noise_map=ag.Array.ones(shape_2d=(3, 4)),
            )
            imaging.image[6] = 4.0

            mask = ag.Mask.manual(
                mask_2d=np.array(
                    [
                        [True, True, True, True],
                        [True, False, False, True],
                        [True, True, True, True],
                    ]
                ),
                pixel_scales=1.0,
                sub_size=1,
            )

            masked_imaging_7x7 = ag.MaskedImaging(
                imaging=imaging, mask=mask, renormalize_psf=False
            )

            # Setup as a ray trace instance, using a light profile for the lens

            g0 = ag.Galaxy(
                redshift=0.5,
                light_profile=MockLightProfile(value=1.0, size=2),
                hyper_galaxy=ag.HyperGalaxy(
                    contribution_factor=1.0, noise_factor=1.0, noise_power=1.0
                ),
                hyper_model_image=ag.Array.ones(shape_2d=(1, 2)),
                hyper_galaxy_image=ag.Array.ones(shape_2d=(1, 2)),
                hyper_minimum_value=0.0,
            )

            plane = ag.Plane(galaxies=[g0])

            fit = ag.FitImaging(masked_imaging=masked_imaging_7x7, plane=plane)

            assert (
                fit.mask
                == np.array(
                    [
                        [True, True, True, True],
                        [True, False, False, True],
                        [True, True, True, True],
                    ]
                )
            ).all()

            assert (
                fit.image.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 5.0, 4.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.noise_map.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 2.0, 2.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.model_image.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 1.0, 4.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.residual_map.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 4.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.normalized_residual_map.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 2.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.chi_squared_map.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 4.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert fit.chi_squared == 4.0
            assert fit.reduced_chi_squared == 4.0 / 2.0
            assert fit.noise_normalization == (2.0 * np.log(2 * np.pi * 2.0 ** 2.0))
            assert fit.log_likelihood == -0.5 * (
                4.0 + 2.0 * np.log(2 * np.pi * 2.0 ** 2.0)
            )

        def test__hyper_image_changes_background_sky__reflected_in_likelihood(self):

            psf = ag.Kernel.manual_2d(
                array=(np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]])),
                pixel_scales=1.0,
            )

            imaging = ag.Imaging(
                image=ag.Array.full(fill_value=4.0, shape_2d=(3, 4)),
                psf=psf,
                noise_map=ag.Array.ones(shape_2d=(3, 4)),
            )
            imaging.image[5] = 5.0

            mask = ag.Mask.manual(
                mask_2d=np.array(
                    [
                        [True, True, True, True],
                        [True, False, False, True],
                        [True, True, True, True],
                    ]
                ),
                pixel_scales=1.0,
                sub_size=1,
            )

            masked_imaging_7x7 = ag.MaskedImaging(imaging=imaging, mask=mask)

            # Setup as a ray trace instance, using a light profile for the lens

            g0 = ag.Galaxy(
                redshift=0.5, light_profile=MockLightProfile(value=1.0, size=2)
            )
            plane = ag.Plane(galaxies=[g0])

            hyper_image_sky = ag.hyper_data.HyperImageSky(sky_scale=1.0)

            fit = ag.FitImaging(
                masked_imaging=masked_imaging_7x7,
                plane=plane,
                hyper_image_sky=hyper_image_sky,
            )

            assert (
                fit.mask
                == np.array(
                    [
                        [True, True, True, True],
                        [True, False, False, True],
                        [True, True, True, True],
                    ]
                )
            ).all()

            assert (
                fit.image.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 6.0, 5.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.noise_map.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 1.0, 1.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.model_image.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 1.0, 1.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.residual_map.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 5.0, 4.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.normalized_residual_map.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 5.0, 4.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.chi_squared_map.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 25.0, 16.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert fit.chi_squared == 41.0
            assert fit.reduced_chi_squared == 41.0 / 2.0
            assert fit.noise_normalization == (2.0 * np.log(2 * np.pi * 1.0 ** 2.0))
            assert fit.log_likelihood == -0.5 * (
                41.0 + 2.0 * np.log(2 * np.pi * 1.0 ** 2.0)
            )

        def test__hyper_background_changes_background_noise_map__reflected_in_likelihood(
            self
        ):

            psf = ag.Kernel.manual_2d(
                array=(np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]])),
                pixel_scales=1.0,
            )

            imaging = ag.Imaging(
                image=5.0 * ag.Array.ones(shape_2d=(3, 4)),
                psf=psf,
                noise_map=ag.Array.ones(shape_2d=(3, 4)),
            )
            imaging.image[6] = 4.0

            mask = ag.Mask.manual(
                mask_2d=np.array(
                    [
                        [True, True, True, True],
                        [True, False, False, True],
                        [True, True, True, True],
                    ]
                ),
                pixel_scales=1.0,
                sub_size=1,
            )

            masked_imaging_7x7 = ag.MaskedImaging(imaging=imaging, mask=mask)

            # Setup as a ray trace instance, using a light profile for the lens

            g0 = ag.Galaxy(
                redshift=0.5, light_profile=MockLightProfile(value=1.0, size=2)
            )
            plane = ag.Plane(galaxies=[g0])

            hyper_background_noise = ag.hyper_data.HyperBackgroundNoise(noise_scale=1.0)

            fit = ag.FitImaging(
                masked_imaging=masked_imaging_7x7,
                plane=plane,
                hyper_background_noise=hyper_background_noise,
            )

            assert (
                fit.mask
                == np.array(
                    [
                        [True, True, True, True],
                        [True, False, False, True],
                        [True, True, True, True],
                    ]
                )
            ).all()

            assert (
                fit.image.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 5.0, 4.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.noise_map.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 2.0, 2.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.model_image.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 1.0, 1.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.residual_map.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 4.0, 3.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.normalized_residual_map.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 2.0, 1.5, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.chi_squared_map.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 4.0, 2.25, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert fit.chi_squared == 6.25
            assert fit.reduced_chi_squared == 6.25 / 2.0
            assert fit.noise_normalization == (2.0 * np.log(2 * np.pi * 2.0 ** 2.0))
            assert fit.log_likelihood == -0.5 * (
                6.25 + 2.0 * np.log(2 * np.pi * 2.0 ** 2.0)
            )

    class TestCompareToManualProfilesOnly:
        def test___all_fit_quantities__no_hyper_methods(self, masked_imaging_7x7):

            g0 = ag.Galaxy(
                redshift=0.5,
                light_profile=ag.lp.EllipticalSersic(intensity=1.0),
                mass_profile=ag.mp.SphericalIsothermal(einstein_radius=1.0),
            )

            g1 = ag.Galaxy(
                redshift=1.0, light_profile=ag.lp.EllipticalSersic(intensity=1.0)
            )

            plane = ag.Plane(redshift=0.75, galaxies=[g0, g1])

            fit = ag.FitImaging(masked_imaging=masked_imaging_7x7, plane=plane)

            assert masked_imaging_7x7.noise_map.in_2d == pytest.approx(
                fit.noise_map.in_2d
            )

            model_image = plane.blurred_profile_image_from_grid_and_convolver(
                grid=masked_imaging_7x7.grid,
                convolver=masked_imaging_7x7.convolver,
                blurring_grid=masked_imaging_7x7.blurring_grid,
            )

            assert model_image.in_2d == pytest.approx(fit.model_image.in_2d)

            residual_map = ag.util.fit.residual_map_from_data_and_model_data(
                data=masked_imaging_7x7.image, model_data=model_image
            )

            assert residual_map.in_2d == pytest.approx(fit.residual_map.in_2d)

            normalized_residual_map = ag.util.fit.normalized_residual_map_from_residual_map_and_noise_map(
                residual_map=residual_map, noise_map=masked_imaging_7x7.noise_map
            )

            assert normalized_residual_map.in_2d == pytest.approx(
                fit.normalized_residual_map.in_2d
            )

            chi_squared_map = ag.util.fit.chi_squared_map_from_residual_map_and_noise_map(
                residual_map=residual_map, noise_map=masked_imaging_7x7.noise_map
            )

            assert chi_squared_map.in_2d == pytest.approx(fit.chi_squared_map.in_2d)

            chi_squared = ag.util.fit.chi_squared_from_chi_squared_map(
                chi_squared_map=chi_squared_map
            )

            noise_normalization = ag.util.fit.noise_normalization_from_noise_map(
                noise_map=masked_imaging_7x7.noise_map
            )

            log_likelihood = ag.util.fit.likelihood_from_chi_squared_and_noise_normalization(
                chi_squared=chi_squared, noise_normalization=noise_normalization
            )

            assert log_likelihood == pytest.approx(fit.log_likelihood, 1e-4)
            assert log_likelihood == fit.figure_of_merit

        def test___fit_galaxy_model_image_dict__corresponds_to_blurred_galaxy_images(
            self, masked_imaging_7x7
        ):
            g0 = ag.Galaxy(
                redshift=0.5,
                light_profile=ag.lp.EllipticalSersic(intensity=1.0),
                mass_profile=ag.mp.SphericalIsothermal(einstein_radius=1.0),
            )
            g1 = ag.Galaxy(
                redshift=1.0, light_profile=ag.lp.EllipticalSersic(intensity=1.0)
            )
            g2 = ag.Galaxy(redshift=1.0)

            plane = ag.Plane(redshift=0.75, galaxies=[g0, g1, g2])

            fit = ag.FitImaging(masked_imaging=masked_imaging_7x7, plane=plane)

            g0_blurred_image = g0.blurred_profile_image_from_grid_and_convolver(
                grid=masked_imaging_7x7.grid,
                blurring_grid=masked_imaging_7x7.blurring_grid,
                convolver=masked_imaging_7x7.convolver,
            )

            g1_blurred_image = g1.blurred_profile_image_from_grid_and_convolver(
                grid=masked_imaging_7x7.grid,
                blurring_grid=masked_imaging_7x7.blurring_grid,
                convolver=masked_imaging_7x7.convolver,
            )

            assert fit.galaxy_model_image_dict[g0] == pytest.approx(
                g0_blurred_image, 1.0e-4
            )
            assert fit.galaxy_model_image_dict[g1] == pytest.approx(
                g1_blurred_image, 1.0e-4
            )
            assert (fit.galaxy_model_image_dict[g2].in_1d == np.zeros(9)).all()

            assert fit.model_image.in_2d == pytest.approx(
                fit.galaxy_model_image_dict[g0].in_2d
                + fit.galaxy_model_image_dict[g1].in_2d,
                1.0e-4,
            )

        def test___all_fit_quantities__including_hyper_methods(
            self, masked_imaging_7x7
        ):

            hyper_image_sky = ag.hyper_data.HyperImageSky(sky_scale=1.0)

            hyper_background_noise = ag.hyper_data.HyperBackgroundNoise(noise_scale=1.0)

            image = hyper_image_sky.hyper_image_from_image(
                image=masked_imaging_7x7.image
            )

            g0 = ag.Galaxy(
                redshift=0.5,
                light_profile=ag.lp.EllipticalSersic(intensity=1.0),
                mass_profile=ag.mp.SphericalIsothermal(einstein_radius=1.0),
                hyper_galaxy=ag.HyperGalaxy(
                    contribution_factor=1.0, noise_factor=1.0, noise_power=1.0
                ),
                hyper_model_image=np.ones(9),
                hyper_galaxy_image=np.ones(9),
                hyper_minimum_value=0.0,
            )
            g1 = ag.Galaxy(
                redshift=1.0, light_profile=ag.lp.EllipticalSersic(intensity=1.0)
            )

            plane = ag.Plane(redshift=0.75, galaxies=[g0, g1])

            fit = ag.FitImaging(
                masked_imaging=masked_imaging_7x7,
                plane=plane,
                hyper_image_sky=hyper_image_sky,
                hyper_background_noise=hyper_background_noise,
            )

            hyper_noise_map_background = hyper_background_noise.hyper_noise_map_from_noise_map(
                noise_map=masked_imaging_7x7.noise_map
            )

            hyper_noise = plane.hyper_noise_map_from_noise_map(
                noise_map=masked_imaging_7x7.noise_map
            )

            hyper_noise_map = hyper_noise_map_background + hyper_noise

            assert hyper_noise_map.in_2d == pytest.approx(fit.noise_map.in_2d)

            model_image = plane.blurred_profile_image_from_grid_and_convolver(
                grid=masked_imaging_7x7.grid,
                convolver=masked_imaging_7x7.convolver,
                blurring_grid=masked_imaging_7x7.blurring_grid,
            )

            assert model_image.in_2d == pytest.approx(fit.model_image.in_2d)

            residual_map = ag.util.fit.residual_map_from_data_and_model_data(
                data=image, model_data=model_image
            )

            assert residual_map.in_2d == pytest.approx(fit.residual_map.in_2d)

            normalized_residual_map = ag.util.fit.normalized_residual_map_from_residual_map_and_noise_map(
                residual_map=residual_map, noise_map=hyper_noise_map
            )

            assert normalized_residual_map.in_2d == pytest.approx(
                fit.normalized_residual_map.in_2d
            )

            chi_squared_map = ag.util.fit.chi_squared_map_from_residual_map_and_noise_map(
                residual_map=residual_map, noise_map=hyper_noise_map
            )

            assert chi_squared_map.in_2d == pytest.approx(fit.chi_squared_map.in_2d)

            chi_squared = ag.util.fit.chi_squared_from_chi_squared_map(
                chi_squared_map=chi_squared_map
            )

            noise_normalization = ag.util.fit.noise_normalization_from_noise_map(
                noise_map=hyper_noise_map
            )

            log_likelihood = ag.util.fit.likelihood_from_chi_squared_and_noise_normalization(
                chi_squared=chi_squared, noise_normalization=noise_normalization
            )

            assert log_likelihood == pytest.approx(fit.log_likelihood, 1e-4)
            assert log_likelihood == fit.figure_of_merit

        def test___blurred_and_model_images_of_galaxies_and_unmasked_blurred_profile_image_properties(
            self, masked_imaging_7x7
        ):

            g0 = ag.Galaxy(
                redshift=0.5,
                light_profile=ag.lp.EllipticalSersic(intensity=1.0),
                mass_profile=ag.mp.SphericalIsothermal(einstein_radius=1.0),
            )

            g1 = ag.Galaxy(
                redshift=1.0, light_profile=ag.lp.EllipticalSersic(intensity=1.0)
            )

            plane = ag.Plane(redshift=0.75, galaxies=[g0, g1])

            fit = ag.FitImaging(masked_imaging=masked_imaging_7x7, plane=plane)

            blurred_profile_images_of_galaxies = plane.blurred_profile_images_of_galaxies_from_grid_and_convolver(
                grid=masked_imaging_7x7.grid,
                convolver=masked_imaging_7x7.convolver,
                blurring_grid=masked_imaging_7x7.blurring_grid,
            )

            assert blurred_profile_images_of_galaxies[0].in_2d == pytest.approx(
                fit.model_images_of_galaxies[0].in_2d, 1.0e-4
            )

            assert blurred_profile_images_of_galaxies[1].in_2d == pytest.approx(
                fit.model_images_of_galaxies[1].in_2d, 1.0e-4
            )

            unmasked_blurred_profile_image = plane.unmasked_blurred_profile_image_from_grid_and_psf(
                grid=masked_imaging_7x7.grid, psf=masked_imaging_7x7.psf
            )

            assert (
                unmasked_blurred_profile_image == fit.unmasked_blurred_profile_image
            ).all()

            unmasked_blurred_profile_image_of_galaxies = plane.unmasked_blurred_profile_image_of_galaxies_from_grid_and_psf(
                grid=masked_imaging_7x7.grid, psf=masked_imaging_7x7.psf
            )

            assert (
                unmasked_blurred_profile_image_of_galaxies[0]
                == fit.unmasked_blurred_profile_image_of_galaxies[0]
            ).all()
            assert (
                unmasked_blurred_profile_image_of_galaxies[1]
                == fit.unmasked_blurred_profile_image_of_galaxies[1]
            ).all()

    class TestCompareToManualInversionOnly:
        def test___all_quantities__no_hyper_methods(self, masked_imaging_7x7):

            pix = ag.pix.Rectangular(shape=(3, 3))
            reg = ag.reg.Constant(coefficient=1.0)

            g0 = ag.Galaxy(redshift=0.5, pixelization=pix, regularization=reg)

            plane = ag.Plane(galaxies=[ag.Galaxy(redshift=0.5), g0])

            fit = ag.FitImaging(masked_imaging=masked_imaging_7x7, plane=plane)

            mapper = pix.mapper_from_grid_and_sparse_grid(
                grid=masked_imaging_7x7.grid, sparse_grid=None
            )
            inversion = inversions.InversionImaging.from_data_mapper_and_regularization(
                mapper=mapper,
                regularization=reg,
                image=masked_imaging_7x7.image,
                noise_map=masked_imaging_7x7.noise_map,
                convolver=masked_imaging_7x7.convolver,
            )

            assert inversion.mapped_reconstructed_image.in_2d == pytest.approx(
                fit.model_image.in_2d, 1.0e-4
            )

            residual_map = ag.util.fit.residual_map_from_data_and_model_data(
                data=masked_imaging_7x7.image,
                model_data=inversion.mapped_reconstructed_image,
            )

            assert residual_map.in_2d == pytest.approx(fit.residual_map.in_2d, 1.0e-4)

            normalized_residual_map = ag.util.fit.normalized_residual_map_from_residual_map_and_noise_map(
                residual_map=residual_map, noise_map=masked_imaging_7x7.noise_map
            )

            assert normalized_residual_map.in_2d == pytest.approx(
                fit.normalized_residual_map.in_2d, 1.0e-4
            )

            chi_squared_map = ag.util.fit.chi_squared_map_from_residual_map_and_noise_map(
                residual_map=residual_map, noise_map=masked_imaging_7x7.noise_map
            )

            assert chi_squared_map.in_2d == pytest.approx(
                fit.chi_squared_map.in_2d, 1.0e-4
            )

            chi_squared = ag.util.fit.chi_squared_from_chi_squared_map(
                chi_squared_map=chi_squared_map
            )

            noise_normalization = ag.util.fit.noise_normalization_from_noise_map(
                noise_map=masked_imaging_7x7.noise_map
            )

            log_likelihood = ag.util.fit.likelihood_from_chi_squared_and_noise_normalization(
                chi_squared=chi_squared, noise_normalization=noise_normalization
            )

            assert log_likelihood == pytest.approx(fit.log_likelihood, 1e-4)

            log_likelihood_with_regularization = ag.util.fit.likelihood_with_regularization_from_inversion_terms(
                chi_squared=chi_squared,
                regularization_term=inversion.regularization_term,
                noise_normalization=noise_normalization,
            )

            assert log_likelihood_with_regularization == pytest.approx(
                fit.log_likelihood_with_regularization, 1e-4
            )

            log_evidence = ag.util.fit.evidence_from_inversion_terms(
                chi_squared=chi_squared,
                regularization_term=inversion.regularization_term,
                log_curvature_regularization_term=inversion.log_det_curvature_reg_matrix_term,
                log_regularization_term=inversion.log_det_regularization_matrix_term,
                noise_normalization=noise_normalization,
            )

            assert log_evidence == fit.log_evidence
            assert log_evidence == fit.figure_of_merit

        def test___fit_galaxy_model_image_dict__has_inversion_mapped_reconstructed_image(
            self, masked_imaging_7x7
        ):
            pix = ag.pix.Rectangular(shape=(3, 3))
            reg = ag.reg.Constant(coefficient=1.0)

            g0 = ag.Galaxy(redshift=0.5)
            g1 = ag.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)

            plane = ag.Plane(redshift=0.75, galaxies=[g0, g1])

            fit = ag.FitImaging(masked_imaging=masked_imaging_7x7, plane=plane)

            mapper = pix.mapper_from_grid_and_sparse_grid(
                grid=masked_imaging_7x7.grid, sparse_grid=None
            )

            inversion = inversions.InversionImaging.from_data_mapper_and_regularization(
                mapper=mapper,
                regularization=reg,
                image=masked_imaging_7x7.image,
                noise_map=masked_imaging_7x7.noise_map,
                convolver=masked_imaging_7x7.convolver,
            )

            assert (fit.galaxy_model_image_dict[g0] == np.zeros(9)).all()

            assert fit.galaxy_model_image_dict[g1].in_2d == pytest.approx(
                inversion.mapped_reconstructed_image.in_2d, 1.0e-4
            )

            assert fit.model_image.in_2d == pytest.approx(
                fit.galaxy_model_image_dict[g1].in_2d, 1.0e-4
            )

        def test___all_fit_quantities__include_hyper_methods(self, masked_imaging_7x7):

            hyper_image_sky = ag.hyper_data.HyperImageSky(sky_scale=1.0)

            hyper_background_noise = ag.hyper_data.HyperBackgroundNoise(noise_scale=1.0)

            image = hyper_image_sky.hyper_image_from_image(
                image=masked_imaging_7x7.image
            )

            hyper_noise_map_background = hyper_background_noise.hyper_noise_map_from_noise_map(
                noise_map=masked_imaging_7x7.noise_map
            )

            pix = ag.pix.Rectangular(shape=(3, 3))
            reg = ag.reg.Constant(coefficient=1.0)

            g0 = ag.Galaxy(
                redshift=0.5,
                pixelization=pix,
                regularization=reg,
                hyper_galaxy=ag.HyperGalaxy(
                    contribution_factor=1.0, noise_factor=1.0, noise_power=1.0
                ),
                hyper_model_image=np.ones(9),
                hyper_galaxy_image=np.ones(9),
                hyper_minimum_value=0.0,
            )

            plane = ag.Plane(galaxies=[ag.Galaxy(redshift=0.5), g0])

            fit = ag.FitImaging(
                masked_imaging=masked_imaging_7x7,
                plane=plane,
                hyper_image_sky=hyper_image_sky,
                hyper_background_noise=hyper_background_noise,
            )

            hyper_noise = plane.hyper_noise_map_from_noise_map(
                noise_map=masked_imaging_7x7.noise_map
            )
            hyper_noise_map = hyper_noise_map_background + hyper_noise

            assert hyper_noise_map.in_2d == pytest.approx(fit.noise_map.in_2d)

            mapper = pix.mapper_from_grid_and_sparse_grid(
                grid=masked_imaging_7x7.grid, inversion_uses_border=False
            )
            inversion = inversions.InversionImaging.from_data_mapper_and_regularization(
                mapper=mapper,
                regularization=reg,
                image=image,
                noise_map=hyper_noise_map,
                convolver=masked_imaging_7x7.convolver,
            )

            assert inversion.mapped_reconstructed_image.in_2d == pytest.approx(
                fit.model_image.in_2d, 1.0e-4
            )

            residual_map = ag.util.fit.residual_map_from_data_and_model_data(
                data=image, model_data=inversion.mapped_reconstructed_image
            )

            assert residual_map.in_2d == pytest.approx(fit.residual_map.in_2d)

            normalized_residual_map = ag.util.fit.normalized_residual_map_from_residual_map_and_noise_map(
                residual_map=residual_map, noise_map=hyper_noise_map
            )

            assert normalized_residual_map.in_2d == pytest.approx(
                fit.normalized_residual_map.in_2d
            )

            chi_squared_map = ag.util.fit.chi_squared_map_from_residual_map_and_noise_map(
                residual_map=residual_map, noise_map=hyper_noise_map
            )

            assert chi_squared_map.in_2d == pytest.approx(fit.chi_squared_map.in_2d)

            chi_squared = ag.util.fit.chi_squared_from_chi_squared_map(
                chi_squared_map=chi_squared_map
            )

            noise_normalization = ag.util.fit.noise_normalization_from_noise_map(
                noise_map=hyper_noise_map
            )

            log_likelihood = ag.util.fit.likelihood_from_chi_squared_and_noise_normalization(
                chi_squared=chi_squared, noise_normalization=noise_normalization
            )

            assert log_likelihood == pytest.approx(fit.log_likelihood, 1e-4)

            log_likelihood_with_regularization = ag.util.fit.likelihood_with_regularization_from_inversion_terms(
                chi_squared=chi_squared,
                regularization_term=inversion.regularization_term,
                noise_normalization=noise_normalization,
            )

            assert log_likelihood_with_regularization == pytest.approx(
                fit.log_likelihood_with_regularization, 1e-4
            )

            log_evidence = ag.util.fit.evidence_from_inversion_terms(
                chi_squared=chi_squared,
                regularization_term=inversion.regularization_term,
                log_curvature_regularization_term=inversion.log_det_curvature_reg_matrix_term,
                log_regularization_term=inversion.log_det_regularization_matrix_term,
                noise_normalization=noise_normalization,
            )

            assert log_evidence == fit.log_evidence
            assert log_evidence == fit.figure_of_merit

        def test___blurred_and_model_images_of_galaxies_and_unmasked_blurred_profile_image_properties(
            self, masked_imaging_7x7
        ):

            pix = ag.pix.Rectangular(shape=(3, 3))
            reg = ag.reg.Constant(coefficient=1.0)

            g0 = ag.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)

            plane = ag.Plane(redshift=0.75, galaxies=[ag.Galaxy(redshift=0.5), g0])

            fit = ag.FitImaging(masked_imaging=masked_imaging_7x7, plane=plane)

            mapper = pix.mapper_from_grid_and_sparse_grid(
                grid=masked_imaging_7x7.grid, inversion_uses_border=False
            )

            inversion = inversions.InversionImaging.from_data_mapper_and_regularization(
                mapper=mapper,
                regularization=reg,
                image=masked_imaging_7x7.image,
                noise_map=masked_imaging_7x7.noise_map,
                convolver=masked_imaging_7x7.convolver,
            )

            assert (fit.model_images_of_galaxies[0].in_2d == np.zeros((7, 7))).all()
            assert inversion.mapped_reconstructed_image.in_2d == pytest.approx(
                fit.model_images_of_galaxies[1].in_2d, 1.0e-4
            )

    class TestCompareToManualProfilesAndInversion:
        def test___all_fit_quantities__no_hyper_methods(self, masked_imaging_7x7):
            galaxy_light = ag.Galaxy(
                redshift=0.5, light_profile=ag.lp.EllipticalSersic(intensity=1.0)
            )

            pix = ag.pix.Rectangular(shape=(3, 3))
            reg = ag.reg.Constant(coefficient=1.0)
            galaxy_pix = ag.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)

            plane = ag.Plane(redshift=0.75, galaxies=[galaxy_light, galaxy_pix])

            fit = ag.FitImaging(masked_imaging=masked_imaging_7x7, plane=plane)

            blurred_profile_image = plane.blurred_profile_image_from_grid_and_convolver(
                grid=masked_imaging_7x7.grid,
                convolver=masked_imaging_7x7.convolver,
                blurring_grid=masked_imaging_7x7.blurring_grid,
            )

            assert blurred_profile_image.in_2d == pytest.approx(
                fit.blurred_profile_image.in_2d
            )

            profile_subtracted_image = masked_imaging_7x7.image - blurred_profile_image

            assert profile_subtracted_image.in_2d == pytest.approx(
                fit.profile_subtracted_image.in_2d
            )

            mapper = pix.mapper_from_grid_and_sparse_grid(
                grid=masked_imaging_7x7.grid, inversion_uses_border=False
            )

            inversion = inversions.InversionImaging.from_data_mapper_and_regularization(
                image=profile_subtracted_image,
                noise_map=masked_imaging_7x7.noise_map,
                convolver=masked_imaging_7x7.convolver,
                mapper=mapper,
                regularization=reg,
            )

            model_image = blurred_profile_image + inversion.mapped_reconstructed_image

            assert model_image.in_2d == pytest.approx(fit.model_image.in_2d)

            residual_map = ag.util.fit.residual_map_from_data_and_model_data(
                data=masked_imaging_7x7.image, model_data=model_image
            )

            assert residual_map.in_2d == pytest.approx(fit.residual_map.in_2d)

            normalized_residual_map = ag.util.fit.normalized_residual_map_from_residual_map_and_noise_map(
                residual_map=residual_map, noise_map=masked_imaging_7x7.noise_map
            )

            assert normalized_residual_map.in_2d == pytest.approx(
                fit.normalized_residual_map.in_2d
            )

            chi_squared_map = ag.util.fit.chi_squared_map_from_residual_map_and_noise_map(
                residual_map=residual_map, noise_map=masked_imaging_7x7.noise_map
            )

            assert chi_squared_map.in_2d == pytest.approx(fit.chi_squared_map.in_2d)

            chi_squared = ag.util.fit.chi_squared_from_chi_squared_map(
                chi_squared_map=chi_squared_map
            )

            noise_normalization = ag.util.fit.noise_normalization_from_noise_map(
                noise_map=masked_imaging_7x7.noise_map
            )

            log_likelihood = ag.util.fit.likelihood_from_chi_squared_and_noise_normalization(
                chi_squared=chi_squared, noise_normalization=noise_normalization
            )

            assert log_likelihood == pytest.approx(fit.log_likelihood, 1e-4)

            log_likelihood_with_regularization = ag.util.fit.likelihood_with_regularization_from_inversion_terms(
                chi_squared=chi_squared,
                regularization_term=inversion.regularization_term,
                noise_normalization=noise_normalization,
            )

            assert log_likelihood_with_regularization == pytest.approx(
                fit.log_likelihood_with_regularization, 1e-4
            )

            log_evidence = ag.util.fit.evidence_from_inversion_terms(
                chi_squared=chi_squared,
                regularization_term=inversion.regularization_term,
                log_curvature_regularization_term=inversion.log_det_curvature_reg_matrix_term,
                log_regularization_term=inversion.log_det_regularization_matrix_term,
                noise_normalization=noise_normalization,
            )

            assert log_evidence == fit.log_evidence
            assert log_evidence == fit.figure_of_merit

        def test___fit_galaxy_model_image_dict__has_blurred_profile_images_and_inversion_mapped_reconstructed_image(
            self, masked_imaging_7x7
        ):

            g0 = ag.Galaxy(
                redshift=0.5, light_profile=ag.lp.EllipticalSersic(intensity=1.0)
            )
            g1 = ag.Galaxy(
                redshift=0.5, light_profile=ag.lp.EllipticalSersic(intensity=2.0)
            )
            g2 = ag.Galaxy(redshift=0.5)

            pix = ag.pix.Rectangular(shape=(3, 3))
            reg = ag.reg.Constant(coefficient=1.0)
            galaxy_pix = ag.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)

            plane = ag.Plane(redshift=0.75, galaxies=[g0, g1, g2, galaxy_pix])

            fit = ag.FitImaging(masked_imaging=masked_imaging_7x7, plane=plane)

            g0_blurred_image = g0.blurred_profile_image_from_grid_and_convolver(
                grid=masked_imaging_7x7.grid,
                convolver=masked_imaging_7x7.convolver,
                blurring_grid=masked_imaging_7x7.blurring_grid,
            )

            g1_blurred_image = g1.blurred_profile_image_from_grid_and_convolver(
                grid=masked_imaging_7x7.grid,
                convolver=masked_imaging_7x7.convolver,
                blurring_grid=masked_imaging_7x7.blurring_grid,
            )

            blurred_profile_image = g0_blurred_image + g1_blurred_image

            profile_subtracted_image = masked_imaging_7x7.image - blurred_profile_image
            mapper = pix.mapper_from_grid_and_sparse_grid(
                grid=masked_imaging_7x7.grid, inversion_uses_border=False
            )

            inversion = inversions.InversionImaging.from_data_mapper_and_regularization(
                image=profile_subtracted_image,
                noise_map=masked_imaging_7x7.noise_map,
                convolver=masked_imaging_7x7.convolver,
                mapper=mapper,
                regularization=reg,
            )

            assert (fit.galaxy_model_image_dict[g2] == np.zeros(9)).all()

            assert fit.galaxy_model_image_dict[g0].in_2d == pytest.approx(
                g0_blurred_image.in_2d, 1.0e-4
            )
            assert fit.galaxy_model_image_dict[g1].in_2d == pytest.approx(
                g1_blurred_image.in_2d, 1.0e-4
            )
            assert fit.galaxy_model_image_dict[galaxy_pix].in_2d == pytest.approx(
                inversion.mapped_reconstructed_image.in_2d, 1.0e-4
            )

            assert fit.model_image.in_2d == pytest.approx(
                fit.galaxy_model_image_dict[g0].in_2d
                + fit.galaxy_model_image_dict[g1].in_2d
                + inversion.mapped_reconstructed_image.in_2d,
                1.0e-4,
            )

        def test___all_fit_quantities__include_hyper_methods(self, masked_imaging_7x7):

            hyper_image_sky = ag.hyper_data.HyperImageSky(sky_scale=1.0)

            hyper_background_noise = ag.hyper_data.HyperBackgroundNoise(noise_scale=1.0)

            image = hyper_image_sky.hyper_image_from_image(
                image=masked_imaging_7x7.image
            )

            hyper_noise_map_background = hyper_background_noise.hyper_noise_map_from_noise_map(
                noise_map=masked_imaging_7x7.noise_map
            )

            galaxy_light = ag.Galaxy(
                redshift=0.5,
                light_profile=ag.lp.EllipticalSersic(intensity=1.0),
                hyper_galaxy=ag.HyperGalaxy(
                    contribution_factor=1.0, noise_factor=1.0, noise_power=1.0
                ),
                hyper_model_image=ag.Array.ones(shape_2d=(3, 3)),
                hyper_galaxy_image=ag.Array.ones(shape_2d=(3, 3)),
                hyper_minimum_value=0.0,
            )

            pix = ag.pix.Rectangular(shape=(3, 3))
            reg = ag.reg.Constant(coefficient=1.0)
            galaxy_pix = ag.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)

            plane = ag.Plane(redshift=0.75, galaxies=[galaxy_light, galaxy_pix])

            fit = ag.FitImaging(
                masked_imaging=masked_imaging_7x7,
                plane=plane,
                hyper_image_sky=hyper_image_sky,
                hyper_background_noise=hyper_background_noise,
            )

            hyper_noise = plane.hyper_noise_map_from_noise_map(
                noise_map=masked_imaging_7x7.noise_map
            )
            hyper_noise_map = hyper_noise_map_background + hyper_noise

            assert hyper_noise_map.in_2d == pytest.approx(fit.noise_map.in_2d, 1.0e-4)

            blurred_profile_image = plane.blurred_profile_image_from_grid_and_convolver(
                grid=masked_imaging_7x7.grid,
                convolver=masked_imaging_7x7.convolver,
                blurring_grid=masked_imaging_7x7.blurring_grid,
            )

            assert blurred_profile_image.in_2d == pytest.approx(
                fit.blurred_profile_image.in_2d
            )

            profile_subtracted_image = image - blurred_profile_image

            assert profile_subtracted_image.in_2d == pytest.approx(
                fit.profile_subtracted_image.in_2d
            )

            mapper = pix.mapper_from_grid_and_sparse_grid(
                grid=masked_imaging_7x7.grid, inversion_uses_border=False
            )

            inversion = inversions.InversionImaging.from_data_mapper_and_regularization(
                image=profile_subtracted_image,
                noise_map=hyper_noise_map,
                convolver=masked_imaging_7x7.convolver,
                mapper=mapper,
                regularization=reg,
            )

            model_image = blurred_profile_image + inversion.mapped_reconstructed_image

            assert model_image.in_2d == pytest.approx(fit.model_image.in_2d, 1.0e-4)

            residual_map = ag.util.fit.residual_map_from_data_and_model_data(
                data=image, model_data=model_image
            )

            assert residual_map.in_2d == pytest.approx(fit.residual_map.in_2d, 1.0e-4)

            normalized_residual_map = ag.util.fit.normalized_residual_map_from_residual_map_and_noise_map(
                residual_map=residual_map, noise_map=hyper_noise_map
            )

            assert normalized_residual_map.in_2d == pytest.approx(
                fit.normalized_residual_map.in_2d, 1.0e-4
            )

            chi_squared_map = ag.util.fit.chi_squared_map_from_residual_map_and_noise_map(
                residual_map=residual_map, noise_map=hyper_noise_map
            )

            assert chi_squared_map.in_2d == pytest.approx(
                fit.chi_squared_map.in_2d, 1.0e-4
            )

            chi_squared = ag.util.fit.chi_squared_from_chi_squared_map(
                chi_squared_map=chi_squared_map
            )

            noise_normalization = ag.util.fit.noise_normalization_from_noise_map(
                noise_map=hyper_noise_map
            )

            log_likelihood = ag.util.fit.likelihood_from_chi_squared_and_noise_normalization(
                chi_squared=chi_squared, noise_normalization=noise_normalization
            )

            assert log_likelihood == pytest.approx(fit.log_likelihood, 1e-4)

            log_likelihood_with_regularization = ag.util.fit.likelihood_with_regularization_from_inversion_terms(
                chi_squared=chi_squared,
                regularization_term=inversion.regularization_term,
                noise_normalization=noise_normalization,
            )

            assert log_likelihood_with_regularization == pytest.approx(
                fit.log_likelihood_with_regularization, 1e-4
            )

            log_evidence = ag.util.fit.evidence_from_inversion_terms(
                chi_squared=chi_squared,
                regularization_term=inversion.regularization_term,
                log_curvature_regularization_term=inversion.log_det_curvature_reg_matrix_term,
                log_regularization_term=inversion.log_det_regularization_matrix_term,
                noise_normalization=noise_normalization,
            )

            assert log_evidence == fit.log_evidence
            assert log_evidence == fit.figure_of_merit

        def test___blurred_and_model_images_of_galaxies_and_unmasked_blurred_profile_image_properties(
            self, masked_imaging_7x7
        ):
            galaxy_light = ag.Galaxy(
                redshift=0.5, light_profile=ag.lp.EllipticalSersic(intensity=1.0)
            )

            pix = ag.pix.Rectangular(shape=(3, 3))
            reg = ag.reg.Constant(coefficient=1.0)
            galaxy_pix = ag.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)

            plane = ag.Plane(redshift=0.75, galaxies=[galaxy_light, galaxy_pix])

            fit = ag.FitImaging(masked_imaging=masked_imaging_7x7, plane=plane)

            blurred_profile_image = plane.blurred_profile_image_from_grid_and_convolver(
                grid=masked_imaging_7x7.grid,
                convolver=masked_imaging_7x7.convolver,
                blurring_grid=masked_imaging_7x7.blurring_grid,
            )

            profile_subtracted_image = masked_imaging_7x7.image - blurred_profile_image

            mapper = pix.mapper_from_grid_and_sparse_grid(
                grid=masked_imaging_7x7.grid, inversion_uses_border=False
            )

            inversion = inversions.InversionImaging.from_data_mapper_and_regularization(
                image=profile_subtracted_image,
                noise_map=masked_imaging_7x7.noise_map,
                convolver=masked_imaging_7x7.convolver,
                mapper=mapper,
                regularization=reg,
            )

            assert blurred_profile_image.in_2d == pytest.approx(
                fit.model_images_of_galaxies[0].in_2d, 1.0e-4
            )
            assert inversion.mapped_reconstructed_image.in_2d == pytest.approx(
                fit.model_images_of_galaxies[1].in_2d, 1.0e-4
            )


class TestFitInterferometer:
    class TestLikelihood:
        def test__1x2_image__1x2_visibilities__simple_fourier_transform(self):
            # The image plane image generated by the galaxy is [1.0, 1.0]

            # Thus the chi squared is 4.0**2.0 + 3.0**2.0 = 25.0

            uv_wavelengths = np.array([[0.0, 0.0]])

            interferometer = ag.Interferometer(
                visibilities=ag.Visibilities.full(fill_value=5.0, shape_1d=(1,)),
                noise_map=ag.Visibilities.ones(shape_1d=(1,)),
                uv_wavelengths=uv_wavelengths,
                primary_beam=None,
            )

            interferometer.visibilities[0, 1] = 4.0

            visibilities_mask = np.full(fill_value=False, shape=(1, 2))

            real_space_mask = ag.Mask.manual(
                mask_2d=np.array(
                    [
                        [True, True, True, True],
                        [True, False, False, True],
                        [True, True, True, True],
                    ]
                ),
                pixel_scales=1.0,
                sub_size=1,
            )

            masked_interferometer = ag.MaskedInterferometer(
                interferometer=interferometer,
                visibilities_mask=visibilities_mask,
                real_space_mask=real_space_mask,
                transformer_class=ag.TransformerDFT,
            )

            # Setup as a ray trace instance, using a light profile for the lens

            g0 = ag.Galaxy(
                redshift=0.5, light_profile=MockLightProfile(value=1.0, size=2)
            )
            plane = ag.Plane(galaxies=[g0])

            fit = ag.FitInterferometer(
                masked_interferometer=masked_interferometer, plane=plane
            )

            assert (fit.visibilities_mask == np.array([False, False])).all()

            assert (fit.visibilities.in_1d == np.array([[5.0, 4.0]])).all()
            assert (fit.noise_map.in_1d == np.array([[1.0, 1.0]])).all()
            assert (fit.model_visibilities.in_1d == np.array([2.0, 0.0])).all()
            assert (fit.residual_map.in_1d == np.array([3.0, 4.0])).all()
            assert (fit.normalized_residual_map.in_1d == np.array([3.0, 4.0])).all()
            assert (fit.chi_squared_map.in_1d == np.array([9.0, 16.0])).all()

            assert fit.chi_squared == 25.0
            assert fit.reduced_chi_squared == 25.0 / 2.0
            assert fit.noise_normalization == (2.0 * np.log(2 * np.pi * 1.0 ** 2.0))
            assert fit.log_likelihood == -0.5 * (
                25.0 + 2.0 * np.log(2 * np.pi * 1.0 ** 2.0)
            )

        def test__hyper_background_changes_background_sky__reflected_in_likelihood(
            self
        ):

            uv_wavelengths = np.array([[1.0, 0.0], [1.0, 1.0], [2.0, 2.0]])

            interferometer = ag.Interferometer(
                visibilities=ag.Visibilities.full(fill_value=5.0, shape_1d=(3,)),
                noise_map=ag.Visibilities.full(fill_value=2.0, shape_1d=(3,)),
                uv_wavelengths=uv_wavelengths,
                primary_beam=None,
            )

            visibilities_mask = np.full(fill_value=False, shape=(1, 2))

            real_space_mask = ag.Mask.manual(
                mask_2d=np.array(
                    [
                        [True, True, True, True, True],
                        [True, False, False, False, True],
                        [True, True, True, True, True],
                    ]
                ),
                pixel_scales=1.0,
                sub_size=1,
            )

            masked_interferometer = ag.MaskedInterferometer(
                interferometer=interferometer,
                visibilities_mask=visibilities_mask,
                real_space_mask=real_space_mask,
            )

            # Setup as a ray trace instance, using a light profile for the lens

            g0 = ag.Galaxy(
                redshift=0.5, light_profile=MockLightProfile(value=1.0, size=2)
            )
            plane = ag.Plane(galaxies=[g0])

            hyper_background_noise = ag.hyper_data.HyperBackgroundNoise(noise_scale=1.0)

            fit = ag.FitInterferometer(
                masked_interferometer=masked_interferometer,
                plane=plane,
                hyper_background_noise=hyper_background_noise,
            )

            assert (
                fit.visibilities.in_1d == np.full(fill_value=5.0, shape=(3, 2))
            ).all()

            assert (fit.noise_map.in_1d == np.full(fill_value=3.0, shape=(3, 2))).all()

    class TestCompareToManualProfilesOnly:
        def test___all_fit_quantities__no_hyper_methods(self, masked_interferometer_7):
            g0 = ag.Galaxy(
                redshift=0.5,
                light_profile=ag.lp.EllipticalSersic(intensity=1.0),
                mass_profile=ag.mp.SphericalIsothermal(einstein_radius=1.0),
            )

            g1 = ag.Galaxy(
                redshift=1.0, light_profile=ag.lp.EllipticalSersic(intensity=1.0)
            )

            plane = ag.Plane(redshift=0.75, galaxies=[g0, g1])

            fit = ag.FitInterferometer(
                masked_interferometer=masked_interferometer_7, plane=plane
            )

            assert masked_interferometer_7.noise_map == pytest.approx(fit.noise_map)

            model_visibilities = plane.profile_visibilities_from_grid_and_transformer(
                grid=masked_interferometer_7.grid,
                transformer=masked_interferometer_7.transformer,
            )

            assert model_visibilities == pytest.approx(fit.model_visibilities, 1e-4)

            residual_map = ag.util.fit.residual_map_from_data_and_model_data(
                data=masked_interferometer_7.visibilities, model_data=model_visibilities
            )

            assert residual_map == pytest.approx(fit.residual_map, 1e-4)

            normalized_residual_map = ag.util.fit.normalized_residual_map_from_residual_map_and_noise_map(
                residual_map=residual_map, noise_map=masked_interferometer_7.noise_map
            )

            assert normalized_residual_map == pytest.approx(
                fit.normalized_residual_map, 1e-4
            )

            chi_squared_map = ag.util.fit.chi_squared_map_from_residual_map_and_noise_map(
                residual_map=residual_map, noise_map=masked_interferometer_7.noise_map
            )

            assert chi_squared_map == pytest.approx(fit.chi_squared_map, 1e-4)

            chi_squared = ag.util.fit.chi_squared_from_chi_squared_map(
                chi_squared_map=fit.chi_squared_map
            )

            noise_normalization = ag.util.fit.noise_normalization_from_noise_map(
                noise_map=masked_interferometer_7.noise_map
            )

            log_likelihood = ag.util.fit.likelihood_from_chi_squared_and_noise_normalization(
                chi_squared=chi_squared, noise_normalization=noise_normalization
            )

            assert log_likelihood == pytest.approx(fit.log_likelihood, 1e-4)
            assert log_likelihood == fit.figure_of_merit

        def test___fit_galaxy_model_image_dict__corresponds_to_profile_galaxy_images(
            self, masked_interferometer_7
        ):
            g0 = ag.Galaxy(
                redshift=0.5,
                light_profile=ag.lp.EllipticalSersic(intensity=1.0),
                mass_profile=ag.mp.SphericalIsothermal(einstein_radius=1.0),
            )
            g1 = ag.Galaxy(
                redshift=1.0, light_profile=ag.lp.EllipticalSersic(intensity=1.0)
            )

            plane = ag.Plane(redshift=0.75, galaxies=[g0, g1])

            fit = ag.FitInterferometer(
                masked_interferometer=masked_interferometer_7, plane=plane
            )

            g0_profile_image = g0.profile_image_from_grid(
                grid=masked_interferometer_7.grid
            )

            g1_profile_image = g1.profile_image_from_grid(
                grid=masked_interferometer_7.grid
            )

            assert fit.galaxy_model_image_dict[g0].in_1d == pytest.approx(
                g0_profile_image, 1.0e-4
            )
            assert fit.galaxy_model_image_dict[g1].in_1d == pytest.approx(
                g1_profile_image, 1.0e-4
            )

        def test___fit_galaxy_visibilities_dict__corresponds_to_galaxy_visibilities(
            self, masked_interferometer_7
        ):
            g0 = ag.Galaxy(
                redshift=0.5,
                light_profile=ag.lp.EllipticalSersic(intensity=1.0),
                mass_profile=ag.mp.SphericalIsothermal(einstein_radius=1.0),
            )
            g1 = ag.Galaxy(
                redshift=1.0, light_profile=ag.lp.EllipticalSersic(intensity=1.0)
            )

            plane = ag.Plane(redshift=0.75, galaxies=[g0, g1])

            fit = ag.FitInterferometer(
                masked_interferometer=masked_interferometer_7, plane=plane
            )

            g0_profile_visibilities = g0.profile_visibilities_from_grid_and_transformer(
                grid=masked_interferometer_7.grid,
                transformer=masked_interferometer_7.transformer,
            )

            g1_profile_visibilities = g1.profile_visibilities_from_grid_and_transformer(
                grid=masked_interferometer_7.grid,
                transformer=masked_interferometer_7.transformer,
            )

            assert fit.galaxy_model_visibilities_dict[g0].in_1d == pytest.approx(
                g0_profile_visibilities, 1.0e-4
            )
            assert fit.galaxy_model_visibilities_dict[g1].in_1d == pytest.approx(
                g1_profile_visibilities, 1.0e-4
            )

            assert fit.model_visibilities.in_1d == pytest.approx(
                fit.galaxy_model_visibilities_dict[g0].in_1d
                + fit.galaxy_model_visibilities_dict[g1].in_1d,
                1.0e-4,
            )

        def test___all_fit_quantities__hyper_background_noise(
            self, masked_interferometer_7
        ):

            hyper_background_noise = ag.hyper_data.HyperBackgroundNoise(noise_scale=1.0)

            hyper_noise_map = hyper_background_noise.hyper_noise_map_from_noise_map(
                noise_map=masked_interferometer_7.noise_map
            )

            g0 = ag.Galaxy(
                redshift=0.5,
                light_profile=ag.lp.EllipticalSersic(intensity=1.0),
                mass_profile=ag.mp.SphericalIsothermal(einstein_radius=1.0),
            )

            g1 = ag.Galaxy(
                redshift=1.0, light_profile=ag.lp.EllipticalSersic(intensity=1.0)
            )

            plane = ag.Plane(redshift=0.75, galaxies=[g0, g1])

            fit = ag.FitInterferometer(
                masked_interferometer=masked_interferometer_7,
                plane=plane,
                hyper_background_noise=hyper_background_noise,
            )

            assert hyper_noise_map.in_1d == pytest.approx(fit.noise_map.in_1d)

    class TestCompareToManualInversionOnly:
        def test___all_fit_quantities__no_hyper_methods(self, masked_interferometer_7):

            pix = ag.pix.Rectangular(shape=(3, 3))
            reg = ag.reg.Constant(coefficient=0.01)

            g0 = ag.Galaxy(redshift=0.5, pixelization=pix, regularization=reg)

            plane = ag.Plane(galaxies=[ag.Galaxy(redshift=0.5), g0])

            fit = ag.FitInterferometer(
                masked_interferometer=masked_interferometer_7, plane=plane
            )

            mapper = pix.mapper_from_grid_and_sparse_grid(
                grid=masked_interferometer_7.grid, sparse_grid=None
            )

            inversion = inversions.InversionInterferometer.from_data_mapper_and_regularization(
                mapper=mapper,
                regularization=reg,
                visibilities=masked_interferometer_7.visibilities,
                noise_map=masked_interferometer_7.noise_map,
                transformer=masked_interferometer_7.transformer,
            )

            assert inversion.mapped_reconstructed_visibilities == pytest.approx(
                fit.model_visibilities, 1.0e-4
            )

            residual_map = ag.util.fit.residual_map_from_data_and_model_data(
                data=masked_interferometer_7.visibilities,
                model_data=inversion.mapped_reconstructed_visibilities,
            )

            assert residual_map.in_1d == pytest.approx(fit.residual_map.in_1d, 1.0e-4)

            normalized_residual_map = ag.util.fit.normalized_residual_map_from_residual_map_and_noise_map(
                residual_map=residual_map, noise_map=masked_interferometer_7.noise_map
            )

            assert normalized_residual_map.in_1d == pytest.approx(
                fit.normalized_residual_map.in_1d, 1.0e-4
            )

            chi_squared_map = ag.util.fit.chi_squared_map_from_residual_map_and_noise_map(
                residual_map=residual_map, noise_map=masked_interferometer_7.noise_map
            )

            assert chi_squared_map.in_1d == pytest.approx(
                fit.chi_squared_map.in_1d, 1.0e-4
            )

            chi_squared = ag.util.fit.chi_squared_from_chi_squared_map(
                chi_squared_map=chi_squared_map
            )

            noise_normalization = ag.util.fit.noise_normalization_from_noise_map(
                noise_map=masked_interferometer_7.noise_map
            )

            log_likelihood = ag.util.fit.likelihood_from_chi_squared_and_noise_normalization(
                chi_squared=chi_squared, noise_normalization=noise_normalization
            )

            assert log_likelihood == pytest.approx(fit.log_likelihood, 1e-4)

            log_likelihood_with_regularization = ag.util.fit.likelihood_with_regularization_from_inversion_terms(
                chi_squared=chi_squared,
                regularization_term=inversion.regularization_term,
                noise_normalization=noise_normalization,
            )

            assert log_likelihood_with_regularization == pytest.approx(
                fit.log_likelihood_with_regularization, 1e-4
            )

            log_evidence = ag.util.fit.evidence_from_inversion_terms(
                chi_squared=chi_squared,
                regularization_term=inversion.regularization_term,
                log_curvature_regularization_term=inversion.log_det_curvature_reg_matrix_term,
                log_regularization_term=inversion.log_det_regularization_matrix_term,
                noise_normalization=noise_normalization,
            )

            assert log_evidence == fit.log_evidence
            assert log_evidence == fit.figure_of_merit

            mapped_reconstructed_image = ag.util.inversion.mapped_reconstructed_data_from_mapping_matrix_and_reconstruction(
                mapping_matrix=fit.inversion.mapper.mapping_matrix,
                reconstruction=fit.inversion.reconstruction,
            )

            assert (
                fit.inversion.mapped_reconstructed_image.in_1d
                == mapped_reconstructed_image
            ).all()

        def test___fit_galaxy_model_image_dict__profile_images_and_inversion_mapped_reconstructed_image(
            self, masked_interferometer_7
        ):
            pix = ag.pix.Rectangular(shape=(3, 3))
            reg = ag.reg.Constant(coefficient=1.0)

            g0 = ag.Galaxy(redshift=0.5)
            g1 = ag.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)

            plane = ag.Plane(redshift=0.75, galaxies=[g0, g1])

            fit = ag.FitInterferometer(
                masked_interferometer=masked_interferometer_7, plane=plane
            )

            mapper = pix.mapper_from_grid_and_sparse_grid(
                grid=masked_interferometer_7.grid, sparse_grid=None
            )

            inversion = inversions.InversionInterferometer.from_data_mapper_and_regularization(
                mapper=mapper,
                regularization=reg,
                visibilities=masked_interferometer_7.visibilities,
                noise_map=masked_interferometer_7.noise_map,
                transformer=masked_interferometer_7.transformer,
            )

            assert (fit.galaxy_model_image_dict[g0].in_2d == np.zeros((7, 7))).all()

            assert fit.galaxy_model_image_dict[g1].in_1d == pytest.approx(
                inversion.mapped_reconstructed_image.in_1d, 1.0e-4
            )

        def test___fit_galaxy_model_visibilities_dict__has_inversion_mapped_reconstructed_visibilities(
            self, masked_interferometer_7
        ):
            pix = ag.pix.Rectangular(shape=(3, 3))
            reg = ag.reg.Constant(coefficient=1.0)

            g0 = ag.Galaxy(redshift=0.5)
            g1 = ag.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)

            plane = ag.Plane(redshift=0.75, galaxies=[g0, g1])

            fit = ag.FitInterferometer(
                masked_interferometer=masked_interferometer_7, plane=plane
            )

            mapper = pix.mapper_from_grid_and_sparse_grid(
                grid=masked_interferometer_7.grid, sparse_grid=None
            )

            inversion = inversions.InversionInterferometer.from_data_mapper_and_regularization(
                mapper=mapper,
                regularization=reg,
                visibilities=masked_interferometer_7.visibilities,
                noise_map=masked_interferometer_7.noise_map,
                transformer=masked_interferometer_7.transformer,
            )

            assert (fit.galaxy_model_visibilities_dict[g0] == np.zeros((7, 2))).all()

            assert fit.galaxy_model_visibilities_dict[g1].in_1d == pytest.approx(
                inversion.mapped_reconstructed_visibilities.in_1d, 1.0e-4
            )

            assert fit.model_visibilities.in_1d == pytest.approx(
                fit.galaxy_model_visibilities_dict[g1].in_1d, 1.0e-4
            )

        def test___all_fit_quantities__hyper_background_noise(
            self, masked_interferometer_7
        ):

            hyper_background_noise = ag.hyper_data.HyperBackgroundNoise(noise_scale=1.0)

            hyper_noise_map = hyper_background_noise.hyper_noise_map_from_noise_map(
                noise_map=masked_interferometer_7.noise_map
            )

            pix = ag.pix.Rectangular(shape=(3, 3))
            reg = ag.reg.Constant(coefficient=0.01)

            g0 = ag.Galaxy(redshift=0.5, pixelization=pix, regularization=reg)

            plane = ag.Plane(galaxies=[ag.Galaxy(redshift=0.5), g0])

            fit = ag.FitInterferometer(
                masked_interferometer=masked_interferometer_7,
                plane=plane,
                hyper_background_noise=hyper_background_noise,
            )

            assert hyper_noise_map.in_1d == pytest.approx(
                fit.inversion.noise_map, 1.0e-4
            )

            assert hyper_noise_map.in_1d == pytest.approx(fit.noise_map.in_1d)

    class TestCompareToManualProfilesAndInversion:
        def test___all_fit_quantities__no_hyper_methods(self, masked_interferometer_7):
            galaxy_light = ag.Galaxy(
                redshift=0.5, light_profile=ag.lp.EllipticalSersic(intensity=1.0)
            )

            pix = ag.pix.Rectangular(shape=(3, 3))
            reg = ag.reg.Constant(coefficient=1.0)
            galaxy_pix = ag.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)

            plane = ag.Plane(redshift=0.75, galaxies=[galaxy_light, galaxy_pix])

            fit = ag.FitInterferometer(
                masked_interferometer=masked_interferometer_7, plane=plane
            )

            profile_visibilities = plane.profile_visibilities_from_grid_and_transformer(
                grid=masked_interferometer_7.grid,
                transformer=masked_interferometer_7.transformer,
            )

            assert profile_visibilities.in_1d == pytest.approx(
                fit.profile_visibilities.in_1d
            )

            profile_subtracted_visibilities = (
                masked_interferometer_7.visibilities - profile_visibilities
            )

            assert profile_subtracted_visibilities.in_1d == pytest.approx(
                fit.profile_subtracted_visibilities.in_1d
            )

            mapper = pix.mapper_from_grid_and_sparse_grid(
                grid=masked_interferometer_7.grid, inversion_uses_border=False
            )

            inversion = inversions.InversionInterferometer.from_data_mapper_and_regularization(
                visibilities=profile_subtracted_visibilities,
                noise_map=masked_interferometer_7.noise_map,
                transformer=masked_interferometer_7.transformer,
                mapper=mapper,
                regularization=reg,
            )

            model_visibilities = (
                profile_visibilities + inversion.mapped_reconstructed_visibilities
            )

            assert model_visibilities.in_1d == pytest.approx(
                fit.model_visibilities.in_1d
            )

            residual_map = ag.util.fit.residual_map_from_data_and_model_data(
                data=masked_interferometer_7.visibilities, model_data=model_visibilities
            )

            assert residual_map.in_1d == pytest.approx(fit.residual_map.in_1d)

            normalized_residual_map = ag.util.fit.normalized_residual_map_from_residual_map_and_noise_map(
                residual_map=residual_map, noise_map=masked_interferometer_7.noise_map
            )

            assert normalized_residual_map.in_1d == pytest.approx(
                fit.normalized_residual_map.in_1d
            )

            chi_squared_map = ag.util.fit.chi_squared_map_from_residual_map_and_noise_map(
                residual_map=residual_map, noise_map=masked_interferometer_7.noise_map
            )

            assert chi_squared_map.in_1d == pytest.approx(fit.chi_squared_map.in_1d)

            chi_squared = ag.util.fit.chi_squared_from_chi_squared_map(
                chi_squared_map=chi_squared_map
            )

            noise_normalization = ag.util.fit.noise_normalization_from_noise_map(
                noise_map=masked_interferometer_7.noise_map
            )

            log_likelihood = ag.util.fit.likelihood_from_chi_squared_and_noise_normalization(
                chi_squared=chi_squared, noise_normalization=noise_normalization
            )

            assert log_likelihood == pytest.approx(fit.log_likelihood, 1e-4)

            log_likelihood_with_regularization = ag.util.fit.likelihood_with_regularization_from_inversion_terms(
                chi_squared=chi_squared,
                regularization_term=inversion.regularization_term,
                noise_normalization=noise_normalization,
            )

            assert log_likelihood_with_regularization == pytest.approx(
                fit.log_likelihood_with_regularization, 1e-4
            )

            log_evidence = ag.util.fit.evidence_from_inversion_terms(
                chi_squared=chi_squared,
                regularization_term=inversion.regularization_term,
                log_curvature_regularization_term=inversion.log_det_curvature_reg_matrix_term,
                log_regularization_term=inversion.log_det_regularization_matrix_term,
                noise_normalization=noise_normalization,
            )

            assert log_evidence == fit.log_evidence
            assert log_evidence == fit.figure_of_merit

            mapped_reconstructed_image = ag.util.inversion.mapped_reconstructed_data_from_mapping_matrix_and_reconstruction(
                mapping_matrix=fit.inversion.mapper.mapping_matrix,
                reconstruction=fit.inversion.reconstruction,
            )

            assert (
                fit.inversion.mapped_reconstructed_image.in_1d
                == mapped_reconstructed_image
            ).all()

        def test___fit_galaxy_model_visibilities_dict__has_profile_image_and_inversion_mapped_reconstructed_image(
            self, masked_interferometer_7
        ):

            g0 = ag.Galaxy(
                redshift=0.5, light_profile=ag.lp.EllipticalSersic(intensity=1.0)
            )
            g1 = ag.Galaxy(
                redshift=0.5, light_profile=ag.lp.EllipticalSersic(intensity=2.0)
            )

            pix = ag.pix.Rectangular(shape=(3, 3))
            reg = ag.reg.Constant(coefficient=1.0)
            galaxy_pix = ag.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)

            plane = ag.Plane(redshift=0.75, galaxies=[g0, g1, galaxy_pix])

            fit = ag.FitInterferometer(
                masked_interferometer=masked_interferometer_7, plane=plane
            )

            g0_visibilities = g0.profile_visibilities_from_grid_and_transformer(
                grid=masked_interferometer_7.grid,
                transformer=masked_interferometer_7.transformer,
            )

            g1_visibilities = g1.profile_visibilities_from_grid_and_transformer(
                grid=masked_interferometer_7.grid,
                transformer=masked_interferometer_7.transformer,
            )

            profile_visibilities = g0_visibilities + g1_visibilities

            profile_subtracted_visibilities = (
                masked_interferometer_7.visibilities - profile_visibilities
            )
            mapper = pix.mapper_from_grid_and_sparse_grid(
                grid=masked_interferometer_7.grid, inversion_uses_border=False
            )

            inversion = inversions.InversionInterferometer.from_data_mapper_and_regularization(
                visibilities=profile_subtracted_visibilities,
                noise_map=masked_interferometer_7.noise_map,
                transformer=masked_interferometer_7.transformer,
                mapper=mapper,
                regularization=reg,
            )

            g0_image = g0.profile_image_from_grid(grid=masked_interferometer_7.grid)

            g1_image = g1.profile_image_from_grid(grid=masked_interferometer_7.grid)

            assert fit.galaxy_model_image_dict[g0].in_1d == pytest.approx(
                g0_image.in_1d, 1.0e-4
            )
            assert fit.galaxy_model_image_dict[g1].in_1d == pytest.approx(
                g1_image.in_1d, 1.0e-4
            )
            assert fit.galaxy_model_image_dict[galaxy_pix].in_1d == pytest.approx(
                inversion.mapped_reconstructed_image.in_1d, 1.0e-4
            )

        def test___fit_galaxy_model_visibilities_dict__has_profile_visibilitiess_and_inversion_mapped_reconstructed_visibilities(
            self, masked_interferometer_7
        ):

            g0 = ag.Galaxy(
                redshift=0.5, light_profile=ag.lp.EllipticalSersic(intensity=1.0)
            )
            g1 = ag.Galaxy(
                redshift=0.5, light_profile=ag.lp.EllipticalSersic(intensity=2.0)
            )
            g2 = ag.Galaxy(redshift=0.5)

            pix = ag.pix.Rectangular(shape=(3, 3))
            reg = ag.reg.Constant(coefficient=1.0)
            galaxy_pix = ag.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)

            plane = ag.Plane(redshift=0.75, galaxies=[g0, g1, g2, galaxy_pix])

            fit = ag.FitInterferometer(
                masked_interferometer=masked_interferometer_7, plane=plane
            )

            g0_visibilities = g0.profile_visibilities_from_grid_and_transformer(
                grid=masked_interferometer_7.grid,
                transformer=masked_interferometer_7.transformer,
            )

            g1_visibilities = g1.profile_visibilities_from_grid_and_transformer(
                grid=masked_interferometer_7.grid,
                transformer=masked_interferometer_7.transformer,
            )

            profile_visibilities = g0_visibilities + g1_visibilities

            profile_subtracted_visibilities = (
                masked_interferometer_7.visibilities - profile_visibilities
            )
            mapper = pix.mapper_from_grid_and_sparse_grid(
                grid=masked_interferometer_7.grid, inversion_uses_border=False
            )

            inversion = inversions.InversionInterferometer.from_data_mapper_and_regularization(
                visibilities=profile_subtracted_visibilities,
                noise_map=masked_interferometer_7.noise_map,
                transformer=masked_interferometer_7.transformer,
                mapper=mapper,
                regularization=reg,
            )

            assert (fit.galaxy_model_visibilities_dict[g2] == np.zeros((7, 2))).all()

            assert fit.galaxy_model_visibilities_dict[g0].in_1d == pytest.approx(
                g0_visibilities.in_1d, 1.0e-4
            )
            assert fit.galaxy_model_visibilities_dict[g1].in_1d == pytest.approx(
                g1_visibilities.in_1d, 1.0e-4
            )
            assert fit.galaxy_model_visibilities_dict[
                galaxy_pix
            ].in_1d == pytest.approx(
                inversion.mapped_reconstructed_visibilities.in_1d, 1.0e-4
            )

            assert fit.model_visibilities.in_1d == pytest.approx(
                fit.galaxy_model_visibilities_dict[g0].in_1d
                + fit.galaxy_model_visibilities_dict[g1].in_1d
                + inversion.mapped_reconstructed_visibilities.in_1d,
                1.0e-4,
            )

        def test___all_fit_quantities__hyper_background_noise(
            self, masked_interferometer_7
        ):

            hyper_background_noise = ag.hyper_data.HyperBackgroundNoise(noise_scale=1.0)

            hyper_noise_map = hyper_background_noise.hyper_noise_map_from_noise_map(
                noise_map=masked_interferometer_7.noise_map
            )

            galaxy_light = ag.Galaxy(
                redshift=0.5, light_profile=ag.lp.EllipticalSersic(intensity=1.0)
            )

            pix = ag.pix.Rectangular(shape=(3, 3))
            reg = ag.reg.Constant(coefficient=1.0)
            galaxy_pix = ag.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)

            plane = ag.Plane(redshift=0.75, galaxies=[galaxy_light, galaxy_pix])

            fit = ag.FitInterferometer(
                masked_interferometer=masked_interferometer_7,
                plane=plane,
                hyper_background_noise=hyper_background_noise,
            )

            assert hyper_noise_map.in_1d == pytest.approx(
                fit.inversion.noise_map, 1.0e-4
            )

            assert hyper_noise_map.in_1d == pytest.approx(fit.noise_map.in_1d)



from autoarray import preprocess
from autoarray.mask.mask import Mask
from autoarray.structures.arrays import Array, Values
from autoarray.structures.grids import Grid, GridRectangular, GridVoronoi, Coordinates
from autoarray.structures.kernel import Kernel
from autoarray.structures.visibilities import Visibilities
from autoarray.structures.arrays import MaskedArray
from autoarray.structures.grids import MaskedGrid
from autoarray.dataset.imaging import Imaging
from autoarray.dataset.interferometer import Interferometer
from autoarray.operators.convolver import Convolver
from autoarray.operators.transformer import TransformerDFT
from autoarray.operators.transformer import TransformerFFT
from autoarray.operators.transformer import TransformerNUFFT
from autoarray.operators.inversion.mappers import mapper as Mapper
from autoarray.operators.inversion.inversions import inversion as Inversion
from autoarray.operators.inversion import pixelizations as pix, regularization as reg
from autoarray import conf

from autogalaxy import aggregator as agg
from autogalaxy.dataset.imaging import MaskedImaging, SimulatorImaging
from autogalaxy.dataset.interferometer import (
    MaskedInterferometer,
    SimulatorInterferometer,
)
from autogalaxy import dimensions as dim
from autogalaxy import util
from autogalaxy.profiles import (
    light_profiles as lp,
    mass_profiles as mp,
    light_and_mass_profiles as lmp,
)
from autogalaxy.galaxy.galaxy import Galaxy, HyperGalaxy, Redshift
from autogalaxy.galaxy.galaxy_data import GalaxyData
from autogalaxy.galaxy.masked_galaxy_data import MaskedGalaxyDataset
from autogalaxy.galaxy.fit_galaxy import FitGalaxy
from autogalaxy.galaxy.galaxy_model import GalaxyModel
from autogalaxy.plane.plane import Plane
from autogalaxy.fit.fit import FitImaging, FitInterferometer
from autogalaxy.hyper import hyper_data
from autogalaxy import plot
from autogalaxy.pipeline import tagging
from autogalaxy.pipeline.phase.abstract import phase
from autogalaxy.pipeline.phase.abstract.phase import AbstractPhase
from autogalaxy.pipeline.phase.extensions import CombinedHyperPhase
from autogalaxy.pipeline.phase.extensions import HyperGalaxyPhase
from autogalaxy.pipeline.phase.extensions.hyper_galaxy_phase import HyperGalaxyPhase
from autogalaxy.pipeline.phase.extensions.hyper_phase import HyperPhase
from autogalaxy.pipeline.phase.extensions.inversion_phase import (
    InversionBackgroundBothPhase,
    InversionBackgroundNoisePhase,
    InversionBackgroundSkyPhase,
    InversionPhase,
    ModelFixingHyperPhase,
)
from autogalaxy.pipeline.phase.abstract.phase import AbstractPhase
from autogalaxy.pipeline.phase.dataset.phase import PhaseDataset
from autogalaxy.pipeline.phase.dataset.meta_dataset import MetaDataset
from autogalaxy.pipeline.phase.imaging.phase import PhaseImaging
from autogalaxy.pipeline.phase.interferometer.phase import PhaseInterferometer
from autogalaxy.pipeline.phase.phase_galaxy import PhaseGalaxy
from autogalaxy.pipeline.pipeline import PipelineDataset
from autogalaxy.pipeline import setup


__version__ = '0.9.2'



import numpy as np
from astropy import cosmology as cosmo
from itertools import count
from scipy.optimize import root_scalar

from autofit.mapper.model_object import ModelObject
from autogalaxy.util import cosmology_util
from autogalaxy import exc
from autogalaxy import dimensions as dim
from autogalaxy import lensing
from autofit.text import formatter
from autoarray.structures import arrays, grids
from autoarray.operators.inversion import pixelizations as pix
from autogalaxy.profiles import light_profiles as lp
from autogalaxy.profiles import mass_profiles as mp
from autogalaxy.profiles.mass_profiles import (
    dark_mass_profiles as dmp,
    stellar_mass_profiles as smp,
)
from copy import deepcopy


def is_light_profile(obj):
    return isinstance(obj, lp.LightProfile)


def is_mass_profile(obj):
    return isinstance(obj, mp.MassProfile)


class Galaxy(ModelObject, lensing.LensingObject):
    """
    @DynamicAttrs
    """

    def __init__(
        self,
        redshift,
        pixelization=None,
        regularization=None,
        hyper_galaxy=None,
        **kwargs,
    ):
        """Class representing a galaxy, which is composed of attributes used for fitting hyper_galaxies (e.g. light profiles, \
        mass profiles, pixelizations, etc.).
        
        All *has_* methods retun *True* if galaxy has that attribute, *False* if not.

        Parameters
        ----------
        redshift: float
            The redshift of the galaxy.
        light_profiles: [lp.LightProfile]
            A list of the galaxy's light profiles.
        mass_profiles: [mp.MassProfile]
            A list of the galaxy's mass profiles.
        hyper_galaxy : HyperGalaxy
            The hyper_galaxies-parameters of the hyper_galaxies-galaxy, which is used for performing a hyper_galaxies-analysis on the noise map.
            
        Attributes
        ----------
        pixelization : inversion.Pixelization
            The pixelization of the galaxy used to reconstruct an observed image using an inversion.
        regularization : inversion.Regularization
            The regularization of the pixel-grid used to reconstruct an observed using an inversion.
        """
        super().__init__()
        self.redshift = redshift

        self.hyper_model_image = None
        self.hyper_galaxy_image = None

        for name, val in kwargs.items():
            setattr(self, name, val)

        self.pixelization = pixelization
        self.regularization = regularization

        if pixelization is not None and regularization is None:
            raise exc.GalaxyException(
                "If the galaxy has a pixelization, it must also have a regularization."
            )
        if pixelization is None and regularization is not None:
            raise exc.GalaxyException(
                "If the galaxy has a regularization, it must also have a pixelization."
            )

        self.hyper_galaxy = hyper_galaxy

    def __hash__(self):
        return self.id

    @property
    def light_profiles(self):
        return [value for value in self.__dict__.values() if is_light_profile(value)]

    @property
    def mass_profiles(self):
        return [value for value in self.__dict__.values() if is_mass_profile(value)]

    @property
    def has_redshift(self):
        return self.redshift is not None

    @property
    def has_pixelization(self):
        return self.pixelization is not None

    @property
    def has_regularization(self):
        return self.regularization is not None

    @property
    def has_hyper_galaxy(self):
        return self.hyper_galaxy is not None

    @property
    def has_light_profile(self):
        return len(self.light_profiles) > 0

    @property
    def has_mass_profile(self):
        return len(self.mass_profiles) > 0

    @property
    def has_only_mass_sheets(self):

        if not self.has_mass_profile:
            return False

        mass_sheet_bools = [
            mass_profile.is_mass_sheet for mass_profile in self.mass_profiles
        ]
        total_mass_sheets = sum(mass_sheet_bools)

        return len(self.mass_profiles) == total_mass_sheets

    @property
    def has_profile(self):
        return len(self.mass_profiles) + len(self.light_profiles) > 0

    @property
    def light_profile_centres(self):
        """Returns the light profile centres of the galaxy as a *Coordinates* object, which structures the centres
        in lists according to which light profile they come from. 
        
        Fo example, if a galaxy has two light profiles, the first with one centre and second with two centres this 
        returns:
        
        [[(y0, x0)], [(y0, x0), (y1, x1)]]

        This is used for visualization, for example plotting the centres of all light profiles colored by their profile.

        NOTE: Currently, no light profiles can have more than one centre (it unlikely one ever will). The structure of 
        the output follows this convention to follow other methods in the *Galaxy* class that return profile 
        attributes."""
        return grids.Coordinates(
            [[light_profile.centre] for light_profile in self.light_profiles]
        )

    @property
    def mass_profile_centres(self):
        """Returns the mass profile centres of the galaxy as a *Coordinates* object, which structures the centres
        in lists according to which mass profile they come from. 

        Fo example, if a galaxy has two mass profiles, the first with one centre and second with two centres this 
        returns:

        [[(y0, x0)], [(y0, x0), (y1, x1)]]

        This is used for visualization, for example plotting the centres of all mass profiles colored by their profile.

        NOTE: Currently, no mass profiles can have more than one centre (it unlikely one ever will). The structure of 
        the output follows this convention to follow other methods in the *Galaxy* class that return profile 
        attributes.

        The centres of mass-sheets are omitted, as their centres are not relevant to lensing calculations."""
        centres = [
            [mass_profile.centre]
            for mass_profile in self.mass_profiles
            if not mass_profile.is_mass_sheet
        ]
        return grids.Coordinates(list(filter(None, centres)))

    @property
    def mass_profile_axis_ratios(self):
        """Returns the mass profile axis-ratios of the galaxy as a *Values* object, which structures the axis-ratios
        in lists according to which mass profile they come from. 

        Fo example, if a galaxy has two mass profiles, the first with one axis-ratio and second with two axis-ratios
        this returns:

        [[axis_ratio_0], [axis_ratio_0, axis_ratio_1]]

        This is used for visualization, for example plotting the axis-ratios of all mass profiles colored by their
        profile.

        """
        return arrays.Values(
            [[mass_profile.axis_ratio] for mass_profile in self.mass_profiles]
        )

    @property
    def mass_profile_phis(self):
        """Returns the mass profile phis of the galaxy as a *Values* object, which structures the phis in lists
        according to which mass profile they come from.

        Fo example, if a galaxy has two mass profiles, the first with one phi and second with two phis this returns:

        [[phi_0], [phi_0, phi_1]]

        This is used for visualization, for example plotting the phis of all mass profiles colored by their profile.

        """
        return arrays.Values(
            [[mass_profile.phi] for mass_profile in self.mass_profiles]
        )

    @property
    def uses_cluster_inversion(self):
        return type(self.pixelization) is pix.VoronoiBrightnessImage

    @property
    def has_stellar_profile(self):
        return len(self.stellar_profiles) > 0

    @property
    def has_dark_profile(self):
        return len(self.dark_profiles) > 0

    @property
    def stellar_profiles(self):
        return [
            profile
            for profile in self.mass_profiles
            if isinstance(profile, smp.StellarProfile)
        ]

    @property
    def dark_profiles(self):
        return [
            profile
            for profile in self.mass_profiles
            if isinstance(profile, dmp.DarkProfile)
        ]

    def stellar_mass_within_circle_in_units(
        self,
        radius: dim.Length,
        unit_mass="angular",
        redshift_source=None,
        cosmology=cosmo.Planck15,
    ):
        if self.has_stellar_profile:
            return sum(
                [
                    profile.mass_within_circle_in_units(
                        radius=radius,
                        unit_mass=unit_mass,
                        redshift_object=self.redshift,
                        redshift_source=redshift_source,
                        cosmology=cosmology,
                    )
                    for profile in self.stellar_profiles
                ]
            )
        else:
            raise exc.GalaxyException(
                "You cannot perform a stellar mass-based calculation on a galaxy which does not have a stellar mass-profile"
            )

    def dark_mass_within_circle_in_units(
        self,
        radius: dim.Length,
        unit_mass="angular",
        redshift_source=None,
        cosmology=cosmo.Planck15,
    ):
        if self.has_dark_profile:
            return sum(
                [
                    profile.mass_within_circle_in_units(
                        radius=radius,
                        unit_mass=unit_mass,
                        redshift_object=self.redshift,
                        redshift_source=redshift_source,
                        cosmology=cosmology,
                    )
                    for profile in self.dark_profiles
                ]
            )
        else:
            raise exc.GalaxyException(
                "You cannot perform a dark mass-based calculation on a galaxy which does not have a dark mass-profile"
            )

    def stellar_fraction_at_radius(self, radius):
        return 1.0 - self.dark_fraction_at_radius(radius=radius)

    def dark_fraction_at_radius(self, radius):

        stellar_mass = self.stellar_mass_within_circle_in_units(radius=radius)
        dark_mass = self.dark_mass_within_circle_in_units(radius=radius)

        return dark_mass / (stellar_mass + dark_mass)

    @property
    def cosmology(self):
        return cosmo.Planck15

    @property
    def arcsec_per_kpc(self):
        return cosmology_util.arcsec_per_kpc_from_redshift_and_cosmology(
            redshift=self.redshift, cosmology=self.cosmology
        )

    @property
    def kpc_per_arcsec(self):
        return 1.0 / self.arcsec_per_kpc

    @property
    def unit_length(self):
        if self.has_light_profile:
            return self.light_profiles[0].unit_length
        elif self.has_mass_profile:
            return self.mass_profiles[0].unit_length
        else:
            return None

    @property
    def unit_luminosity(self):
        if self.has_light_profile:
            return self.light_profiles[0].unit_luminosity
        elif self.has_mass_profile:
            return self.mass_profiles[0].unit_luminosity
        else:
            return None

    @property
    def unit_mass(self):
        if self.has_mass_profile:
            return self.mass_profiles[0].unit_mass
        else:
            return None

    def __repr__(self):
        string = "Redshift: {}".format(self.redshift)
        if self.pixelization:
            string += "\nPixelization:\n{}".format(str(self.pixelization))
        if self.regularization:
            string += "\nRegularization:\n{}".format(str(self.regularization))
        if self.hyper_galaxy:
            string += "\nHyper Galaxy:\n{}".format(str(self.hyper_galaxy))
        if self.light_profiles:
            string += "\nLight Profiles:\n{}".format(
                "\n".join(map(str, self.light_profiles))
            )
        if self.mass_profiles:
            string += "\nMass Profiles:\n{}".format(
                "\n".join(map(str, self.mass_profiles))
            )
        return string

    def __eq__(self, other):
        return all(
            (
                isinstance(other, Galaxy),
                self.pixelization == other.pixelization,
                self.redshift == other.redshift,
                self.hyper_galaxy == other.hyper_galaxy,
                self.light_profiles == other.light_profiles,
                self.mass_profiles == other.mass_profiles,
            )
        )

    def new_object_with_units_converted(
        self,
        unit_length=None,
        unit_luminosity=None,
        unit_mass=None,
        kpc_per_arcsec=None,
        exposure_time=None,
        critical_surface_density=None,
    ):

        new_dict = {
            key: value.new_object_with_units_converted(
                unit_length=unit_length,
                unit_luminosity=unit_luminosity,
                unit_mass=unit_mass,
                kpc_per_arcsec=kpc_per_arcsec,
                exposure_time=exposure_time,
                critical_surface_density=critical_surface_density,
            )
            if is_light_profile(value) or is_mass_profile(value)
            else value
            for key, value in self.__dict__.items()
        }

        return self.__class__(**new_dict)

    @grids.grid_like_to_numpy
    def profile_image_from_grid(self, grid):
        """Calculate the summed image of all of the galaxy's light profiles using a grid of Cartesian (y,x) \
        coordinates.
        
        If the galaxy has no light profiles, a grid of zeros is returned.
        
        See *profiles.light_profiles* for a description of how light profile image are computed.

        Parameters
        ----------
        grid : grid_like
            The (y, x) coordinates in the original reference frame of the grid.

        """
        if self.has_light_profile:
            return sum(
                map(lambda p: p.profile_image_from_grid(grid=grid), self.light_profiles)
            )
        return np.zeros((grid.shape[0],))

    def blurred_profile_image_from_grid_and_psf(self, grid, psf, blurring_grid=None):

        profile_image = self.profile_image_from_grid(grid=grid)

        blurring_image = self.profile_image_from_grid(grid=blurring_grid)

        return psf.convolved_array_from_array_2d_and_mask(
            array_2d=profile_image.in_2d_binned + blurring_image.in_2d_binned,
            mask=grid.mask,
        )

    def blurred_profile_image_from_grid_and_convolver(
        self, grid, convolver, blurring_grid
    ):

        profile_image = self.profile_image_from_grid(grid=grid)

        blurring_image = self.profile_image_from_grid(grid=blurring_grid)

        return convolver.convolved_image_from_image_and_blurring_image(
            image=profile_image.in_1d_binned, blurring_image=blurring_image.in_1d_binned
        )

    def profile_visibilities_from_grid_and_transformer(self, grid, transformer):

        profile_image = self.profile_image_from_grid(grid=grid)

        return transformer.visibilities_from_image(image=profile_image.in_1d_binned)

    def luminosity_within_circle_in_units(
        self,
        radius: dim.Length,
        unit_luminosity="eps",
        exposure_time=None,
        cosmology=cosmo.Planck15,
        **kwargs,
    ):
        """Compute the total luminosity of the galaxy's light profiles within a circle of specified radius.

        See *light_profiles.luminosity_within_circle* for details of how this is performed.

        Parameters
        ----------
        radius : float
            The radius of the circle to compute the dimensionless mass within.
        unit_luminosity : str
            The unit_label the luminosity is returned in (eps | counts).
        exposure_time : float
            The exposure time of the observation, which converts luminosity from electrons per second unit_label to counts.
        """
        if self.has_light_profile:
            return sum(
                map(
                    lambda p: p.luminosity_within_circle_in_units(
                        radius=radius,
                        unit_luminosity=unit_luminosity,
                        redshift_object=self.redshift,
                        exposure_time=exposure_time,
                        cosmology=cosmology,
                        kwargs=kwargs,
                    ),
                    self.light_profiles,
                )
            )
        return None

    @grids.grid_like_to_numpy
    def convergence_from_grid(self, grid):
        """Compute the summed convergence of the galaxy's mass profiles using a grid of Cartesian (y,x) coordinates.

        If the galaxy has no mass profiles, a grid of zeros is returned.
        
        See *profiles.mass_profiles* module for details of how this is performed.

        The *grid_like_to_numpy* decorator reshapes the NumPy arrays the convergence is outputted on. See \
        *aa.grid_like_to_numpy* for a description of the output.

        Parameters
        ----------
        grid : grid_like
            The (y, x) coordinates in the original reference frame of the grid.

        """
        if self.has_mass_profile:
            return sum(
                map(lambda p: p.convergence_from_grid(grid=grid), self.mass_profiles)
            )
        return np.zeros((grid.shape[0],))

    @grids.grid_like_to_numpy
    def potential_from_grid(self, grid):
        """Compute the summed gravitational potential of the galaxy's mass profiles \
        using a grid of Cartesian (y,x) coordinates.

        If the galaxy has no mass profiles, a grid of zeros is returned.

        See *profiles.mass_profiles* module for details of how this is performed.

        The *grid_like_to_numpy* decorator reshapes the NumPy arrays the convergence is outputted on. See \
        *aa.grid_like_to_numpy* for a description of the output.

        Parameters
        ----------
        grid : grid_like
            The (y, x) coordinates in the original reference frame of the grid.

        """
        if self.has_mass_profile:
            return sum(
                map(lambda p: p.potential_from_grid(grid=grid), self.mass_profiles)
            )
        return np.zeros((grid.shape[0],))

    @grids.grid_like_to_numpy
    def deflections_from_grid(self, grid):
        """Compute the summed (y,x) deflection angles of the galaxy's mass profiles \
        using a grid of Cartesian (y,x) coordinates.

        If the galaxy has no mass profiles, two grid of zeros are returned.

        See *profiles.mass_profiles* module for details of how this is performed.

        Parameters
        ----------
        grid : grid_like
            The (y, x) coordinates in the original reference frame of the grid.
        """
        if self.has_mass_profile:
            return sum(
                map(lambda p: p.deflections_from_grid(grid=grid), self.mass_profiles)
            )
        return np.zeros((grid.shape[0], 2))

    def mass_within_circle_in_units(
        self,
        radius: dim.Length,
        unit_mass="angular",
        redshift_source=None,
        cosmology=cosmo.Planck15,
    ):
        """ Integrate the mass profiles's convergence profile to compute the total mass within a circle of \
        specified radius. This is centred on the mass profile.

        The following unit_label for mass can be specified and output:

        - Dimensionless angular unit_label (default) - 'angular'.
        - Solar masses - 'angular' (multiplies the angular mass by the critical surface mass density).

        Parameters
        ----------
        radius : dim.Length
            The radius of the circle to compute the dimensionless mass within.
        unit_mass : str
            The unit_label the mass is returned in (angular | angular).
        critical_surface_density : float or None
            The critical surface mass density of the strong lens configuration, which converts mass from angulalr \
            unit_label to phsical unit_label (e.g. solar masses).
        """
        if self.has_mass_profile:
            return sum(
                map(
                    lambda p: p.mass_within_circle_in_units(
                        radius=radius,
                        unit_mass=unit_mass,
                        redshift_object=self.redshift,
                        redshift_source=redshift_source,
                        cosmology=cosmology,
                    ),
                    self.mass_profiles,
                )
            )
        else:
            raise exc.GalaxyException(
                "You cannot perform a mass-based calculation on a galaxy which does not have a mass-profile"
            )

    @property
    def contribution_map(self):
        """Compute the contribution map of a galaxy, which represents the fraction of
        flux in each pixel that the galaxy is attributed to contain, hyper to the
        *contribution_factor* hyper_galaxies-parameter.

        This is computed by dividing that galaxy's flux by the total flux in that \
        pixel and then scaling by the maximum flux such that the contribution map \
        ranges between 0 and 1.

        Parameters
        -----------
        hyper_model_image : ndarray
            The best-fit model image to the observed image from a previous analysis
            phase. This provides the total light attributed to each image pixel by the
            model.
        hyper_galaxy_image : ndarray
            A model image of the galaxy (from light profiles or an inversion) from a
            previous analysis phase.
        """
        return self.hyper_galaxy.contribution_map_from_hyper_images(
            hyper_model_image=self.hyper_model_image,
            hyper_galaxy_image=self.hyper_galaxy_image,
        )

    def summarize_in_units(
        self,
        radii,
        whitespace=80,
        unit_length="arcsec",
        unit_luminosity="eps",
        unit_mass="solMass",
        redshift_source=None,
        cosmology=cosmo.Planck15,
        **kwargs,
    ):

        if hasattr(self, "name"):
            summary = ["Galaxy = {}\n".format(self.name)]
            prefix_galaxy = self.name + "_"
        else:
            summary = ["Galaxy\n"]
            prefix_galaxy = ""

        summary += [
            formatter.label_and_value_string(
                label=prefix_galaxy + "redshift",
                value=self.redshift,
                whitespace=whitespace,
            )
        ]

        if self.has_light_profile:
            summary += self.summarize_light_profiles_in_units(
                whitespace=whitespace,
                prefix=prefix_galaxy,
                radii=radii,
                unit_length=unit_length,
                unit_luminosity=unit_luminosity,
                redshift_source=redshift_source,
                cosmology=cosmology,
                kwargs=kwargs,
            )

        if self.has_mass_profile:
            summary += self.summarize_mass_profiles_in_units(
                whitespace=whitespace,
                prefix=prefix_galaxy,
                radii=radii,
                unit_length=unit_length,
                unit_mass=unit_mass,
                redshift_source=redshift_source,
                cosmology=cosmology,
                kwargs=kwargs,
            )

        return summary

    def summarize_light_profiles_in_units(
        self,
        radii,
        whitespace=80,
        prefix="",
        unit_length="arcsec",
        unit_luminosity="eps",
        redshift_source=None,
        cosmology=cosmo.Planck15,
        **kwargs,
    ):

        summary = ["\nGALAXY LIGHT\n\n"]

        for radius in radii:
            luminosity = self.luminosity_within_circle_in_units(
                unit_luminosity=unit_luminosity,
                radius=radius,
                redshift_source=redshift_source,
                cosmology=cosmology,
                kwargs=kwargs,
            )

            summary += [
                formatter.within_radius_label_value_and_unit_string(
                    prefix=prefix + "luminosity",
                    radius=radius,
                    unit_length=unit_length,
                    value=luminosity,
                    unit_value=unit_luminosity,
                    whitespace=whitespace,
                )
            ]

        summary.append("\nLIGHT PROFILES:\n\n")

        for light_profile in self.light_profiles:
            summary += light_profile.summarize_in_units(
                radii=radii,
                whitespace=whitespace,
                unit_length=unit_length,
                unit_luminosity=unit_luminosity,
                redshift_profile=self.redshift,
                redshift_source=redshift_source,
                cosmology=cosmology,
                kwargs=kwargs,
            )

            summary += "\n"

        return summary

    def summarize_mass_profiles_in_units(
        self,
        radii,
        whitespace=80,
        prefix="",
        unit_length="arcsec",
        unit_mass="solMass",
        redshift_source=None,
        cosmology=cosmo.Planck15,
        **kwargs,
    ):

        summary = ["\nGALAXY MASS\n\n"]

        einstein_radius = self.einstein_radius_in_units(
            unit_length=unit_length, cosmology=cosmology
        )

        summary += [
            formatter.label_value_and_unit_string(
                label=prefix + "einstein_radius",
                value=einstein_radius,
                unit=unit_length,
                whitespace=whitespace,
            )
        ]

        einstein_mass = self.einstein_mass_in_units(
            unit_mass=unit_mass, redshift_source=redshift_source, cosmology=cosmology
        )

        summary += [
            formatter.label_value_and_unit_string(
                label=prefix + "einstein_mass",
                value=einstein_mass,
                unit=unit_mass,
                whitespace=whitespace,
            )
        ]

        for radius in radii:
            mass = self.mass_within_circle_in_units(
                unit_mass=unit_mass,
                radius=radius,
                redshift_source=redshift_source,
                cosmology=cosmology,
            )

            summary += [
                formatter.within_radius_label_value_and_unit_string(
                    prefix=prefix + "mass",
                    radius=radius,
                    unit_length=unit_length,
                    value=mass,
                    unit_value=unit_mass,
                    whitespace=whitespace,
                )
            ]

        summary += ["\nMASS PROFILES:\n\n"]

        for mass_profile in self.mass_profiles:
            summary += mass_profile.summarize_in_units(
                radii=radii,
                whitespace=whitespace,
                unit_length=unit_length,
                unit_mass=unit_mass,
                redshift_profile=self.redshift,
                redshift_source=redshift_source,
                cosmology=cosmology,
            )

            summary += "\n"

        return summary


class HyperGalaxy:
    _ids = count()

    def __init__(self, contribution_factor=0.0, noise_factor=0.0, noise_power=1.0):
        """ If a *Galaxy* is given a *HyperGalaxy* as an attribute, the noise map in \
        the regions of the image that the galaxy is located will be hyper, to prevent \
        over-fitting of the galaxy.
        
        This is performed by first computing the hyper_galaxies-galaxy's 'contribution-map', \
        which determines the fraction of flux in every pixel of the image that can be \
        associated with this particular hyper_galaxies-galaxy. This is computed using \
        hyper_galaxies-hyper_galaxies set (e.g. fitting.fit_data.FitDataHyper), which includes  best-fit \
        unblurred_image_1d of the galaxy's light from a previous analysis phase.
         
        The *HyperGalaxy* class contains the hyper_galaxies-parameters which are associated \
        with this galaxy for scaling the noise map.
        
        Parameters
        -----------
        contribution_factor : float
            Factor that adjusts how much of the galaxy's light is attributed to the
            contribution map.
        noise_factor : float
            Factor by which the noise map is increased in the regions of the galaxy's
            contribution map.
        noise_power : float
            The power to which the contribution map is raised when scaling the
            noise map.
        """
        self.contribution_factor = contribution_factor
        self.noise_factor = noise_factor
        self.noise_power = noise_power

        self.component_number = next(self._ids)

    def contribution_map_from_hyper_images(self, hyper_model_image, hyper_galaxy_image):
        """Compute the contribution map of a galaxy, which represents the fraction of
        flux in each pixel that the galaxy is attributed to contain, hyper to the
        *contribution_factor* hyper_galaxies-parameter.

        This is computed by dividing that galaxy's flux by the total flux in that \
        pixel and then scaling by the maximum flux such that the contribution map \
        ranges between 0 and 1.

        Parameters
        -----------
        hyper_model_image : ndarray
            The best-fit model image to the observed image from a previous analysis
            phase. This provides the total light attributed to each image pixel by the
            model.
        hyper_galaxy_image : ndarray
            A model image of the galaxy (from light profiles or an inversion) from a
            previous analysis phase.
        """
        contribution_map = np.divide(
            hyper_galaxy_image, np.add(hyper_model_image, self.contribution_factor)
        )
        return np.divide(contribution_map, np.max(contribution_map))

    def hyper_noise_map_from_hyper_images_and_noise_map(
        self, hyper_model_image, hyper_galaxy_image, noise_map
    ):
        contribution_map = self.contribution_map_from_hyper_images(
            hyper_model_image=hyper_model_image, hyper_galaxy_image=hyper_galaxy_image
        )
        return self.hyper_noise_map_from_contribution_map(
            noise_map=noise_map, contribution_map=contribution_map
        )

    def hyper_noise_map_from_contribution_map(self, noise_map, contribution_map):
        """Compute a hyper galaxy hyper_galaxies noise map from a baseline noise map.

        This uses the galaxy contribution map and the *noise_factor* and *noise_power*
        hyper_galaxies-parameters.

        Parameters
        -----------
        noise_map : ndarray
            The observed noise map (before scaling).
        contribution_map : ndarray
            The galaxy contribution map.
        """
        return self.noise_factor * (noise_map * contribution_map) ** self.noise_power

    def __eq__(self, other):
        if isinstance(other, HyperGalaxy):
            return (
                self.contribution_factor == other.contribution_factor
                and self.noise_factor == other.noise_factor
                and self.noise_power == other.noise_power
            )
        return False

    def __str__(self):
        return "\n".join(["{}: {}".format(k, v) for k, v in self.__dict__.items()])


class Redshift(float):
    def __new__(cls, redshift):
        # noinspection PyArgumentList
        return float.__new__(cls, redshift)

    def __init__(self, redshift):
        float.__init__(redshift)



import autofit as af
from autoarray.exc import InversionException
from autogalaxy.galaxy import galaxy as g
from autofit.exc import FitException
from autogalaxy.fit import fit
from autogalaxy.pipeline import visualizer
from autogalaxy.pipeline.phase.dataset import analysis as analysis_data


class Analysis(analysis_data.Analysis):
    def __init__(self, masked_interferometer, cosmology, image_path=None, results=None):

        super(Analysis, self).__init__(cosmology=cosmology, results=results)

        self.visualizer = visualizer.PhaseInterferometerVisualizer(
            masked_dataset=masked_interferometer, image_path=image_path
        )

        self.visualizer.visualize_hyper_images(
            hyper_galaxy_image_path_dict=self.hyper_galaxy_image_path_dict,
            hyper_model_image=self.hyper_model_image,
        )

        self.masked_dataset = masked_interferometer

        result = analysis_data.last_result_with_use_as_hyper_dataset(results=results)

        if result is not None:

            self.hyper_galaxy_visibilities_path_dict = (
                result.hyper_galaxy_visibilities_path_dict
            )

            self.hyper_model_visibilities = result.hyper_model_visibilities

        else:

            self.hyper_galaxy_visibilities_path_dict = None
            self.hyper_model_visibilities = None

    @property
    def masked_interferometer(self):
        return self.masked_dataset

    def log_likelihood_function(self, instance):
        """
        Determine the fit of a lens galaxy and source galaxy to the masked_interferometer in this lens.

        Parameters
        ----------
        instance
            A model instance with attributes

        Returns
        -------
        fit : Fit
            A fractional value indicating how well this model fit and the model masked_interferometer itself
        """

        self.associate_hyper_images(instance=instance)
        plane = self.plane_for_instance(instance=instance)

        # self.masked_dataset.check_inversion_pixels_are_below_limit_via_tracer(
        #     plane=plane
        # )

        hyper_background_noise = self.hyper_background_noise_for_instance(
            instance=instance
        )

        try:
            fit = self.masked_interferometer_fit_for_plane(
                plane=plane, hyper_background_noise=hyper_background_noise
            )

            return fit.figure_of_merit
        except InversionException as e:
            raise FitException from e

    def associate_hyper_visibilities(
        self, instance: af.ModelInstance
    ) -> af.ModelInstance:
        """
        Takes visibilities from the last result, if there is one, and associates them with galaxies in this phase
        where full-path galaxy names match.

        If the galaxy collection has a different name then an association is not made.

        e.g.
        galaxies.lens will match with:
            galaxies.lens
        but not with:
            galaxies.lens
            galaxies.source

        Parameters
        ----------
        instance
            A model instance with 0 or more galaxies in its tree

        Returns
        -------
        instance
           The input instance with visibilities associated with galaxies where possible.
        """
        if self.hyper_galaxy_visibilities_path_dict is not None:
            for galaxy_path, galaxy in instance.path_instance_tuples_for_class(
                g.Galaxy
            ):
                if galaxy_path in self.hyper_galaxy_visibilities_path_dict:
                    galaxy.hyper_model_visibilities = self.hyper_model_visibilities
                    galaxy.hyper_galaxy_visibilities = self.hyper_galaxy_visibilities_path_dict[
                        galaxy_path
                    ]

        return instance

    def masked_interferometer_fit_for_plane(self, plane, hyper_background_noise):

        return fit.FitInterferometer(
            masked_interferometer=self.masked_dataset,
            plane=plane,
            hyper_background_noise=hyper_background_noise,
        )

    def visualize(self, instance, during_analysis):

        self.associate_hyper_images(instance=instance)
        plane = self.plane_for_instance(instance=instance)
        hyper_background_noise = self.hyper_background_noise_for_instance(
            instance=instance
        )

        fit = self.masked_interferometer_fit_for_plane(
            plane=plane, hyper_background_noise=hyper_background_noise
        )

        if plane.has_mass_profile:

            visualizer = self.visualizer.new_visualizer_with_preloaded_critical_curves_and_caustics(
                preloaded_critical_curves=plane.critical_curves,
                preloaded_caustics=plane.caustics,
            )

        else:

            visualizer = self.visualizer

        visualizer.visualize_fit(fit=fit, during_analysis=during_analysis)



import autofit as af
import numpy as np
from astropy import cosmology as cosmo
from scipy.integrate import quad

from autoarray.structures import grids
from autogalaxy.util import cosmology_util
from autogalaxy import dimensions as dim
from autofit.text import formatter
from autogalaxy.profiles import geometry_profiles


class LightProfile:
    """Mixin class that implements functions common to all light profiles"""

    def profile_image_from_grid_radii(self, grid_radii):
        """
        Abstract method for obtaining intensity at on a grid of radii.

        Parameters
        ----------
        grid_radii : float
            The radial distance from the centre of the profile. for each coordinate on the grid.
        """
        raise NotImplementedError("intensity_at_radius should be overridden")

    # noinspection PyMethodMayBeStatic
    def profile_image_from_grid(self, grid, grid_radial_minimum=None):
        """
        Abstract method for obtaining intensity at a grid of Cartesian (y,x) coordinates.

        Parameters
        ----------
        grid : grid_like
            The (y, x) coordinates in the original reference frame of the grid.
        Returns
        -------
        intensity : ndarray
            The value of intensity at the given radius
        """
        raise NotImplementedError("intensity_from_grid should be overridden")

    def luminosity_within_circle_in_units(
        self,
        radius: dim.Length,
        unit_luminosity="eps",
        exposure_time=None,
        redshift_profile=None,
        cosmology=cosmo.Planck15,
        **kwargs
    ):
        raise NotImplementedError()

    def summarize_in_units(
        self,
        radii,
        unit_length="arcsec",
        unit_luminosity="eps",
        exposure_time=None,
        redshift_profile=None,
        cosmology=cosmo.Planck15,
        **kwargs
    ):
        return ["Light Profile = {}\n".format(self.__class__.__name__)]


# noinspection PyAbstractClass
class EllipticalLightProfile(geometry_profiles.EllipticalProfile, LightProfile):
    """Generic class for an elliptical light profiles"""

    @af.map_types
    def __init__(
        self,
        centre: dim.Position = (0.0, 0.0),
        axis_ratio: float = 1.0,
        phi: float = 0.0,
        intensity: dim.Luminosity = 0.1,
    ):
        """  Abstract class for an elliptical light-profile.

        Parameters
        ----------
        centre : (float, float)
            The (y,x) arc-second coordinates of the profile centre.
        axis_ratio : float
            Ratio of light profiles ellipse's minor and major axes (b/a)
        phi : float
            Rotational angle of profiles ellipse counter-clockwise from positive x-axis
        """
        super(EllipticalLightProfile, self).__init__(
            centre=centre, axis_ratio=axis_ratio, phi=phi
        )
        self.intensity = intensity

    @property
    def light_profile_centres(self):
        return grids.Coordinates([self.centre])

    def blurred_profile_image_from_grid_and_psf(self, grid, psf, blurring_grid):
        """Evaluate the light profile image on an input *Grid* of coordinates and then convolve it with a PSF.

        The *Grid* may be masked, in which case values outside but near the edge of the mask will convolve light into
        the mask. A blurring grid is therefore required, which evaluates the image on pixels on the mask edge such that
        their light is blurred into it by the PSF.

        The grid and blurring_grid must be a *Grid* objects so the evaluated image can be mapped to a uniform 2D array
        and binned up for convolution. They therefore cannot be *Coordinates* objects.

        Parameters
        ----------
        grid : Grid
            The (y, x) coordinates in the original reference frame of the grid.
        psf : aa.Kernel
            The PSF the evaluated light profile image is convolved with.
        blurring_grid : Grid
            The (y,x) coordinates neighboring the (masked) grid whose light is blurred into the image.

        """
        profile_image = self.profile_image_from_grid(grid=grid)

        blurring_image = self.profile_image_from_grid(grid=blurring_grid)

        return psf.convolved_array_from_array_2d_and_mask(
            array_2d=profile_image.in_2d_binned + blurring_image.in_2d_binned,
            mask=grid.mask,
        )

    def blurred_profile_image_from_grid_and_convolver(
        self, grid, convolver, blurring_grid
    ):
        """Evaluate the light profile image on an input *Grid* of coordinates and then convolve it with a PSF using a
        *Convolver* object.

        The *Grid* may be masked, in which case values outside but near the edge of the mask will convolve light into
        the mask. A blurring grid is therefore required, which evaluates the image on pixels on the mask edge such that
        their light is blurred into it by the Convolver.

        The grid and blurring_grid must be a *Grid* objects so the evaluated image can be mapped to a uniform 2D array
        and binned up for convolution. They therefore cannot be *Coordinates* objects.

        Parameters
        ----------
        grid : Grid
            The (y, x) coordinates in the original reference frame of the grid.
        Convolver : aa.Convolver
            The Convolver object used to blur the PSF.
        blurring_grid : Grid
            The (y,x) coordinates neighboring the (masked) grid whose light is blurred into the image.

        """
        profile_image = self.profile_image_from_grid(grid=grid)

        blurring_image = self.profile_image_from_grid(grid=blurring_grid)

        return convolver.convolved_image_from_image_and_blurring_image(
            image=profile_image.in_1d_binned, blurring_image=blurring_image.in_1d_binned
        )

    def profile_visibilities_from_grid_and_transformer(self, grid, transformer):

        profile_image = self.profile_image_from_grid(grid=grid)

        return transformer.visibilities_from_image(image=profile_image.in_1d_binned)

    def luminosity_within_circle_in_units(
        self,
        radius: dim.Length,
        unit_luminosity="eps",
        exposure_time=None,
        redshift_object=None,
        cosmology=cosmo.Planck15,
        **kwargs
    ):
        """Integrate the light profile to compute the total luminosity within a circle of specified radius. This is \
        centred on the light profile's centre.

        The following unit_label for mass can be specified and output:

        - Electrons per second (default) - 'eps'.
        - Counts - 'counts' (multiplies the luminosity in electrons per second by the exposure time).

        Parameters
        ----------
        radius : float
            The radius of the circle to compute the dimensionless mass within.
        unit_luminosity : str
            The unit_label the luminosity is returned in (eps | counts).
        exposure_time : float or None
            The exposure time of the observation, which converts luminosity from electrons per second unit_label to counts.
        """

        if not hasattr(radius, "unit_length"):
            radius = dim.Length(value=radius, unit_length="arcsec")

        if self.unit_length is not radius.unit_length:

            kpc_per_arcsec = cosmology_util.kpc_per_arcsec_from_redshift_and_cosmology(
                redshift=redshift_object, cosmology=cosmology
            )

            radius = radius.convert(
                unit_length=self.unit_length, kpc_per_arcsec=kpc_per_arcsec
            )

        luminosity = dim.Luminosity(
            value=quad(self.luminosity_integral, a=0.0, b=radius)[0],
            unit_luminosity=self.unit_luminosity,
        )
        return luminosity.convert(
            unit_luminosity=unit_luminosity, exposure_time=exposure_time
        )

    def luminosity_integral(self, x):
        """Routine to integrate the luminosity of an elliptical light profile.

        The axis ratio is set to 1.0 for computing the luminosity within a circle"""
        return 2 * np.pi * x * self.profile_image_from_grid_radii(x)

    def summarize_in_units(
        self,
        radii,
        prefix="",
        unit_length="arcsec",
        unit_luminosity="eps",
        exposure_time=None,
        redshift_profile=None,
        cosmology=cosmo.Planck15,
        whitespace=80,
        **kwargs
    ):
        summary = super().summarize_in_units(
            radii=radii,
            unit_length=unit_length,
            unit_luminosity=unit_luminosity,
            exposure_time=exposure_time,
            redshift_profile=redshift_profile,
            cosmology=cosmology,
            kwargs=kwargs,
        )

        for radius in radii:
            luminosity = self.luminosity_within_circle_in_units(
                unit_luminosity=unit_luminosity,
                radius=radius,
                redshift_object=redshift_profile,
                exposure_time=exposure_time,
                cosmology=cosmology,
                kwargs=kwargs,
            )

            summary += [
                formatter.within_radius_label_value_and_unit_string(
                    prefix=prefix + "luminosity",
                    radius=radius,
                    unit_length=unit_length,
                    value=luminosity,
                    unit_value=unit_luminosity,
                    whitespace=whitespace,
                )
            ]

        return summary


class EllipticalGaussian(EllipticalLightProfile):
    @af.map_types
    def __init__(
        self,
        centre: dim.Position = (0.0, 0.0),
        axis_ratio: float = 1.0,
        phi: float = 0.0,
        intensity: dim.Luminosity = 0.1,
        sigma: dim.Length = 0.01,
    ):
        """ The elliptical Gaussian light profile.

        Parameters
        ----------
        centre : (float, float)
            The (y,x) arc-second coordinates of the profile centre.
        axis_ratio : float
            Ratio of light profiles ellipse's minor and major axes (b/a).
        phi : float
            Rotation angle of light profile counter-clockwise from positive x-axis.
        intensity : float
            Overall intensity normalisation of the light profiles (electrons per second).
        sigma : float
            The sigma value of the Gaussian, correspodning to ~ 1 / sqrt(2 log(2)) the full width half maximum.
        """
        super(EllipticalGaussian, self).__init__(
            centre=centre, axis_ratio=axis_ratio, phi=phi, intensity=intensity
        )
        self.sigma = sigma

    def profile_image_from_grid_radii(self, grid_radii):
        """Calculate the intensity of the Gaussian light profile on a grid of radial coordinates.

        Parameters
        ----------
        grid_radii : float
            The radial distance from the centre of the profile. for each coordinate on the grid.
        """
        return np.multiply(
            np.divide(self.intensity, self.sigma * np.sqrt(2.0 * np.pi)),
            np.exp(-0.5 * np.square(np.divide(grid_radii, self.sigma))),
        )

    @grids.grid_like_to_numpy
    @grids.transform_grid
    @grids.move_grid_to_radial_minimum
    def profile_image_from_grid(self, grid, grid_radial_minimum=None):
        """
        Calculate the intensity of the light profile on a grid of Cartesian (y,x) coordinates.

        If the coordinates have not been transformed to the profile's geometry, this is performed automatically.

        Parameters
        ----------
        grid : grid_like
            The (y, x) coordinates in the original reference frame of the grid.
        """
        return self.profile_image_from_grid_radii(self.grid_to_elliptical_radii(grid))


class SphericalGaussian(EllipticalGaussian):
    @af.map_types
    def __init__(
        self,
        centre: dim.Position = (0.0, 0.0),
        intensity: dim.Luminosity = 0.1,
        sigma: dim.Length = 0.01,
    ):
        """ The spherical Gaussian light profile.

        Parameters
        ----------
        centre : (float, float)
            The (y,x) arc-second coordinates of the profile centre.
        intensity : float_
            Overall intensity normalisation of the light profiles (electrons per second).
        sigma : float
            The sigma value of the Gaussian, correspodning to ~ 1 / sqrt(2 log(2)) the full width half maximum.
        """
        super(SphericalGaussian, self).__init__(
            centre=centre, axis_ratio=1.0, phi=0.0, intensity=intensity, sigma=sigma
        )


class AbstractEllipticalSersic(EllipticalLightProfile):
    @af.map_types
    def __init__(
        self,
        centre: dim.Position = (0.0, 0.0),
        axis_ratio: float = 1.0,
        phi: float = 0.0,
        intensity: dim.Luminosity = 0.1,
        effective_radius: dim.Length = 0.6,
        sersic_index: float = 4.0,
    ):
        """ Abstract base class for an elliptical Sersic light profile, used for computing its effective radius and
        Sersic instance.

        Parameters
        ----------
        centre : (float, float)
            The (y,x) arc-second coordinates of the profile centre.
        axis_ratio : float
            Ratio of light profiles ellipse's minor and major axes (b/a)
        phi : float
            Rotational angle of profiles ellipse counter-clockwise from positive x-axis
        intensity : float
            Overall intensity normalisation in the light profiles (electrons per second)
        effective_radius : float
            The circular radius containing half the light of this model_mapper
        sersic_index : Int
            Controls the concentration of the of the profile (lower value -> less concentrated, \
            higher value -> more concentrated).
        """
        super(AbstractEllipticalSersic, self).__init__(
            centre=centre, axis_ratio=axis_ratio, phi=phi, intensity=intensity
        )
        self.effective_radius = effective_radius
        self.sersic_index = sersic_index

    def new_profile_with_units_distance_converted(
        self, units_distance, kpc_per_arcsec=None
    ):
        self.units_distance = units_distance
        self.centre = self.centre.convert(
            unit_distance=units_distance, kpc_per_arcsec=kpc_per_arcsec
        )
        self.effective_radius = self.effective_radius.convert(
            unit_distance=units_distance, kpc_per_arcsec=kpc_per_arcsec
        )
        return self

    @property
    def elliptical_effective_radius(self):
        """The effective_radius of a Sersic light profile is defined as the circular effective radius. This is the \
        radius within which a circular aperture contains half the profiles's total integrated light. For elliptical \
        systems, this won't robustly capture the light profile's elliptical shape.

        The elliptical effective radius instead describes the major-axis radius of the ellipse containing \
        half the light, and may be more appropriate for highly flattened systems like disk galaxies."""
        return self.effective_radius / np.sqrt(self.axis_ratio)

    @property
    def sersic_constant(self):
        """ A parameter derived from Sersic index which ensures that effective radius contains 50% of the profile's
        total integrated light.
        """
        return (
            (2 * self.sersic_index)
            - (1.0 / 3.0)
            + (4.0 / (405.0 * self.sersic_index))
            + (46.0 / (25515.0 * self.sersic_index ** 2))
            + (131.0 / (1148175.0 * self.sersic_index ** 3))
            - (2194697.0 / (30690717750.0 * self.sersic_index ** 4))
        )

    def intensity_at_radius(self, radius):
        """ Compute the intensity of the profile at a given radius.

        Parameters
        ----------
        radius : float
            The distance from the centre of the profile.
        """
        return self.intensity * np.exp(
            -self.sersic_constant
            * (((radius / self.effective_radius) ** (1.0 / self.sersic_index)) - 1)
        )


class EllipticalSersic(AbstractEllipticalSersic, EllipticalLightProfile):
    @af.map_types
    def __init__(
        self,
        centre: dim.Position = (0.0, 0.0),
        axis_ratio: float = 1.0,
        phi: float = 0.0,
        intensity: dim.Luminosity = 0.1,
        effective_radius: dim.Length = 0.6,
        sersic_index: float = 4.0,
    ):
        """ The elliptical Sersic light profile.

        Parameters
        ----------
        centre : (float, float)
            The (y,x) arc-second coordinates of the profile centre.
        axis_ratio : float
            Ratio of light profiles ellipse's minor and major axes (b/a).
        phi : float
            Rotation angle of light profile counter-clockwise from positive x-axis.
        intensity : float
            Overall intensity normalisation of the light profiles (electrons per second).
        effective_radius : float
            The circular radius containing half the light of this profile.
        sersic_index : Int
            Controls the concentration of the of the profile (lower value -> less concentrated, \
            higher value -> more concentrated).
        """
        super(EllipticalSersic, self).__init__(
            centre=centre,
            axis_ratio=axis_ratio,
            phi=phi,
            intensity=intensity,
            effective_radius=effective_radius,
            sersic_index=sersic_index,
        )

    def profile_image_from_grid_radii(self, grid_radii):
        """
        Calculate the intensity of the Sersic light profile on a grid of radial coordinates.

        Parameters
        ----------
        grid_radii : float
            The radial distance from the centre of the profile. for each coordinate on the grid.
        """
        np.seterr(all="ignore")
        return np.multiply(
            self.intensity,
            np.exp(
                np.multiply(
                    -self.sersic_constant,
                    np.add(
                        np.power(
                            np.divide(grid_radii, self.effective_radius),
                            1.0 / self.sersic_index,
                        ),
                        -1,
                    ),
                )
            ),
        )

    @grids.grid_like_to_numpy
    @grids.transform_grid
    @grids.move_grid_to_radial_minimum
    def profile_image_from_grid(self, grid, grid_radial_minimum=None):
        """ Calculate the intensity of the light profile on a grid of Cartesian (y,x) coordinates.

        If the coordinates have not been transformed to the profile's geometry, this is performed automatically.

        Parameters
        ----------
        grid : grid_like
            The (y, x) coordinates in the original reference frame of the grid.
        """
        return self.profile_image_from_grid_radii(self.grid_to_eccentric_radii(grid))


class SphericalSersic(EllipticalSersic):
    @af.map_types
    def __init__(
        self,
        centre: dim.Position = (0.0, 0.0),
        intensity: dim.Luminosity = 0.1,
        effective_radius: dim.Length = 0.6,
        sersic_index: float = 4.0,
    ):
        """ The spherical Sersic light profile.

        Parameters
        ----------
        centre : (float, float)
            The (y,x) arc-second coordinates of the profile centre.
        intensity : float
            Overall intensity normalisation of the light profiles (electrons per second).
        effective_radius : float
            The circular radius containing half the light of this profile.
        sersic_index : Int
            Controls the concentration of the of the light profile.
        """
        super(SphericalSersic, self).__init__(
            centre=centre,
            axis_ratio=1.0,
            phi=0.0,
            intensity=intensity,
            effective_radius=effective_radius,
            sersic_index=sersic_index,
        )


class EllipticalExponential(EllipticalSersic):
    @af.map_types
    def __init__(
        self,
        centre: dim.Position = (0.0, 0.0),
        axis_ratio: float = 1.0,
        phi: float = 0.0,
        intensity: dim.Luminosity = 0.1,
        effective_radius: dim.Length = 0.6,
    ):
        """ The elliptical exponential profile.

        This is a subset of the elliptical Sersic profile, specific to the case that sersic_index = 1.0.

        Parameters
        ----------
        centre : (float, float)
            The (y,x) arc-second centre of the light profile.
        axis_ratio : float
            Ratio of light profiles ellipse's minor and major axes (b/a).
        phi : float
            Rotation angle of light profile counter-clockwise from positive x-axis.
        intensity : float
            Overall intensity normalisation of the light profiles (electrons per second).
        effective_radius : float
            The circular radius containing half the light of this profile.
        """
        super(EllipticalExponential, self).__init__(
            centre=centre,
            axis_ratio=axis_ratio,
            phi=phi,
            intensity=intensity,
            effective_radius=effective_radius,
            sersic_index=1.0,
        )


class SphericalExponential(EllipticalExponential):
    @af.map_types
    def __init__(
        self,
        centre: dim.Position = (0.0, 0.0),
        intensity: dim.Luminosity = 0.1,
        effective_radius: dim.Length = 0.6,
    ):
        """ The spherical exponential profile.

        This is a subset of the elliptical Sersic profile, specific to the case that sersic_index = 1.0.

        Parameters
        ----------
        centre : (float, float)
            The (y,x) arc-second coordinates of the profile centre.
        intensity : float
            Overall intensity normalisation of the light profiles (electrons per second).
        effective_radius : float
            The circular radius containing half the light of this profile.
        """
        super(SphericalExponential, self).__init__(
            centre=centre,
            axis_ratio=1.0,
            phi=0.0,
            intensity=intensity,
            effective_radius=effective_radius,
        )


class EllipticalDevVaucouleurs(EllipticalSersic):
    @af.map_types
    def __init__(
        self,
        centre: dim.Position = (0.0, 0.0),
        axis_ratio: float = 1.0,
        phi: float = 0.0,
        intensity: dim.Luminosity = 0.1,
        effective_radius: dim.Length = 0.6,
    ):
        """ The elliptical Dev Vaucouleurs light profile.

        This is a subset of the elliptical Sersic profile, specific to the case that sersic_index = 4.0.

        Parameters
        ----------
        centre : (float, float)
            The (y,x) arc-second coordinates of the profile centre.
        axis_ratio : float
            Ratio of light profiles ellipse's minor and major axes (b/a).
        phi : float
            Rotation angle of light profile counter-clockwise from positive x-axis.
        intensity : float
            Overall intensity normalisation of the light profiles (electrons per second).
        effective_radius : float
            The circular radius containing half the light of this profile.
        """
        super(EllipticalDevVaucouleurs, self).__init__(
            centre=centre,
            axis_ratio=axis_ratio,
            phi=phi,
            intensity=intensity,
            effective_radius=effective_radius,
            sersic_index=4.0,
        )


class SphericalDevVaucouleurs(EllipticalDevVaucouleurs):
    @af.map_types
    def __init__(
        self,
        centre: dim.Position = (0.0, 0.0),
        intensity: dim.Luminosity = 0.1,
        effective_radius: dim.Length = 0.6,
    ):
        """ The spherical Dev Vaucouleurs light profile.

        This is a subset of the elliptical Sersic profile, specific to the case that sersic_index = 1.0.

        Parameters
        ----------
        centre : (float, float)
            The (y,x) arc-second coordinates of the profile centre.
        intensity : float
            Overall intensity normalisation of the light profiles (electrons per second).
        effective_radius : float
            The circular radius containing half the light of this profile.
        """
        super(SphericalDevVaucouleurs, self).__init__(
            centre=centre,
            axis_ratio=1.0,
            phi=0.0,
            intensity=intensity,
            effective_radius=effective_radius,
        )


class EllipticalCoreSersic(EllipticalSersic):
    @af.map_types
    def __init__(
        self,
        centre: dim.Position = (0.0, 0.0),
        axis_ratio: float = 1.0,
        phi: float = 0.0,
        intensity: dim.Luminosity = 0.1,
        effective_radius: dim.Length = 0.6,
        sersic_index: float = 4.0,
        radius_break: dim.Length = 0.01,
        intensity_break: dim.Luminosity = 0.05,
        gamma: float = 0.25,
        alpha: float = 3.0,
    ):
        """ The elliptical cored-Sersic light profile.

        Parameters
        ----------
        centre : (float, float)
            The (y,x) arc-second coordinates of the profile centre.
        axis_ratio : float
            Ratio of light profiles ellipse's minor and major axes (b/a).
        phi : float
            Rotation angle of light profile counter-clockwise from positive x-axis.
        intensity : float
            Overall intensity normalisation of the light profiles (electrons per second).
        effective_radius : float
            The circular radius containing half the light of this profile.
        sersic_index : Int
            Controls the concentration of the of the profile (lower value -> less concentrated, \
            higher value -> more concentrated).
        radius_break : Float
            The break radius separating the inner power-law (with logarithmic slope gamma) and outer Sersic function.
        intensity_break : Float
            The intensity at the break radius.
        gamma : Float
            The logarithmic power-law slope of the inner core profiles
        alpha :
            Controls the sharpness of the transition between the inner core / outer Sersic profiles.
        """
        super(EllipticalCoreSersic, self).__init__(
            centre=centre,
            axis_ratio=axis_ratio,
            phi=phi,
            intensity=intensity,
            effective_radius=effective_radius,
            sersic_index=sersic_index,
        )
        self.radius_break = radius_break
        self.intensity_break = intensity_break
        self.alpha = alpha
        self.gamma = gamma

    def new_profile_with_units_distance_converted(
        self, units_distance, kpc_per_arcsec=None
    ):
        self.units_distance = units_distance
        self.centre = self.centre.convert(
            unit_distance=units_distance, kpc_per_arcsec=kpc_per_arcsec
        )
        self.effective_radius = self.effective_radius.convert(
            unit_distance=units_distance, kpc_per_arcsec=kpc_per_arcsec
        )
        self.radius_break = self.radius_break.convert(
            unit_distance=units_distance, kpc_per_arcsec=kpc_per_arcsec
        )
        return self

    @property
    def intensity_prime(self):
        """Overall intensity normalisation in the rescaled Core-Sersic light profiles (electrons per second)"""
        return (
            self.intensity_break
            * (2.0 ** (-self.gamma / self.alpha))
            * np.exp(
                self.sersic_constant
                * (
                    ((2.0 ** (1.0 / self.alpha)) * self.radius_break)
                    / self.effective_radius
                )
                ** (1.0 / self.sersic_index)
            )
        )

    def profile_image_from_grid_radii(self, grid_radii):
        """Calculate the intensity of the cored-Sersic light profile on a grid of radial coordinates.

        Parameters
        ----------
        grid_radii : float
            The radial distance from the centre of the profile. for each coordinate on the grid.
        """
        return np.multiply(
            np.multiply(
                self.intensity_prime,
                np.power(
                    np.add(
                        1,
                        np.power(np.divide(self.radius_break, grid_radii), self.alpha),
                    ),
                    (self.gamma / self.alpha),
                ),
            ),
            np.exp(
                np.multiply(
                    -self.sersic_constant,
                    (
                        np.power(
                            np.divide(
                                np.add(
                                    np.power(grid_radii, self.alpha),
                                    (self.radius_break ** self.alpha),
                                ),
                                (self.effective_radius ** self.alpha),
                            ),
                            (1.0 / (self.alpha * self.sersic_index)),
                        )
                    ),
                )
            ),
        )


class SphericalCoreSersic(EllipticalCoreSersic):
    @af.map_types
    def __init__(
        self,
        centre: dim.Position = (0.0, 0.0),
        intensity: dim.Luminosity = 0.1,
        effective_radius: dim.Length = 0.6,
        sersic_index: float = 4.0,
        radius_break: dim.Length = 0.01,
        intensity_break: dim.Luminosity = 0.05,
        gamma: float = 0.25,
        alpha: float = 3.0,
    ):
        """ The elliptical cored-Sersic light profile.

        Parameters
        ----------
        centre : (float, float)
            The (y,x) arc-second coordinates of the profile centre.
        intensity : float
            Overall intensity normalisation of the light profiles (electrons per second).
        effective_radius : float
            The circular radius containing half the light of this profile.
        sersic_index : Int
            Controls the concentration of the of the profile (lower value -> less concentrated, \
            higher value -> more concentrated).
        radius_break : Float
            The break radius separating the inner power-law (with logarithmic slope gamma) and outer Sersic function.
        intensity_break : Float
            The intensity at the break radius.
        gamma : Float
            The logarithmic power-law slope of the inner core profiles
        alpha :
            Controls the sharpness of the transition between the inner core / outer Sersic profiles.
        """
        super(SphericalCoreSersic, self).__init__(
            centre=centre,
            axis_ratio=1.0,
            phi=0.0,
            intensity=intensity,
            effective_radius=effective_radius,
            sersic_index=sersic_index,
            radius_break=radius_break,
            intensity_break=intensity_break,
            gamma=gamma,
            alpha=alpha,
        )
        self.radius_break = radius_break
        self.intensity_break = intensity_break
        self.alpha = alpha
        self.gamma = gamma



from astropy import cosmology as cosmo

import autofit as af
from autofit.tools.phase import Dataset
from autogalaxy.pipeline.phase import abstract
from autogalaxy.pipeline.phase import extensions
from autogalaxy.pipeline.phase.dataset.result import Result


class PhaseDataset(abstract.AbstractPhase):
    galaxies = af.PhaseProperty("galaxies")

    Result = Result

    @af.convert_paths
    def __init__(
        self,
        paths,
        galaxies=None,
        non_linear_class=af.MultiNest,
        cosmology=cosmo.Planck15,
    ):
        """

        A phase in an lens pipeline. Uses the set non_linear optimizer to try to fit models and hyper_galaxies
        passed to it.

        Parameters
        ----------
        non_linear_class: class
            The class of a non_linear optimizer
        """

        super().__init__(paths, non_linear_class=non_linear_class)
        self.galaxies = galaxies or []
        self.cosmology = cosmology
        self.use_as_hyper_dataset = False
        self.is_hyper_phase = False

    def run(self, dataset: Dataset, mask, results=None, info=None):
        """
        Run this phase.

        Parameters
        ----------
        mask: Mask
            The default masks passed in by the pipeline
        results: autofit.tools.pipeline.ResultsCollection
            An object describing the results of the last phase or None if no phase has been executed
        dataset: scaled_array.ScaledSquarePixelArray
            An masked_imaging that has been masked

        Returns
        -------
        result: AbstractPhase.Result
            A result object comprising the best fit model and other hyper_galaxies.
        """
        self.save_model_info()
        self.save_metadata(dataset=dataset)
        self.save_dataset(dataset=dataset)
        self.save_mask(mask)
        self.save_meta_dataset(meta_dataset=self.meta_dataset)
        self.save_info(info=info)

        self.model = self.model.populate(results)

        results = results or af.ResultsCollection()

        analysis = self.make_analysis(dataset=dataset, mask=mask, results=results)

        phase_attributes = self.make_phase_attributes(analysis=analysis)
        self.save_phase_attributes(phase_attributes=phase_attributes)

        self.customize_priors(results)
        self.assert_and_save_pickle()

        result = self.run_analysis(analysis)

        return self.make_result(result=result, analysis=analysis)

    def make_analysis(self, dataset, mask, results=None):
        """
        Create an lens object. Also calls the prior passing and masked_imaging modifying functions to allow child
        classes to change the behaviour of the phase.

        Parameters
        ----------
        mask: Mask
            The default masks passed in by the pipeline
        dataset: im.Imaging
            An masked_imaging that has been masked
        results: autofit.tools.pipeline.ResultsCollection
            The result from the previous phase

        Returns
        -------
        lens : Analysis
            An lens object that the non-linear optimizer calls to determine the fit of a set of values
        """
        raise NotImplementedError()

    def extend_with_inversion_phase(self, non_linear_class=af.MultiNest):
        return extensions.InversionPhase(phase=self, non_linear_class=non_linear_class)

    def extend_with_multiple_hyper_phases(
        self,
        hyper_galaxy=False,
        inversion=False,
        include_background_sky=False,
        include_background_noise=False,
        hyper_galaxy_phase_first=False,
    ):

        self.use_as_hyper_dataset = True

        hyper_phase_classes = []

        if self.meta_dataset.has_pixelization and inversion:
            if not include_background_sky and not include_background_noise:
                hyper_phase_classes.append(extensions.InversionPhase)
            elif include_background_sky and not include_background_noise:
                hyper_phase_classes.append(extensions.InversionBackgroundSkyPhase)
            elif not include_background_sky and include_background_noise:
                hyper_phase_classes.append(extensions.InversionBackgroundNoisePhase)
            else:
                hyper_phase_classes.append(extensions.InversionBackgroundBothPhase)

        if hyper_galaxy:
            if not include_background_sky and not include_background_noise:
                hyper_phase_classes.append(
                    extensions.hyper_galaxy_phase.HyperGalaxyPhase
                )
            elif include_background_sky and not include_background_noise:
                hyper_phase_classes.append(
                    extensions.hyper_galaxy_phase.HyperGalaxyBackgroundSkyPhase
                )
            elif not include_background_sky and include_background_noise:
                hyper_phase_classes.append(
                    extensions.hyper_galaxy_phase.HyperGalaxyBackgroundNoisePhase
                )
            else:
                hyper_phase_classes.append(
                    extensions.hyper_galaxy_phase.HyperGalaxyBackgroundBothPhase
                )

        if hyper_galaxy_phase_first:
            if inversion and hyper_galaxy:
                hyper_phase_classes = [cls for cls in reversed(hyper_phase_classes)]

        if len(hyper_phase_classes) == 0:
            return self
        else:
            return extensions.CombinedHyperPhase(
                phase=self, hyper_phase_classes=hyper_phase_classes
            )
